import os
from logging import getLogger

from ubiquerg import is_url

from ..utils import load_yaml
from .const import PROP_KEY, SAMPLES_KEY

_LOGGER = getLogger(__name__)


def preprocess_schema(schema_dict: dict) -> dict:
    """
    Preprocess schema before validation for user's convenience.

    Preprocessing includes:
    - renaming 'samples' to '_samples' since in the peppy.Project object
        _samples attribute holds the list of peppy.Samples objects.
    - adding array of strings entry for every string specified to accommodate
        subsamples in peppy.Project

    Args:
        schema_dict: Schema dictionary to preprocess

    Returns:
        Preprocessed schema
    """
    _LOGGER.debug(f"schema ori: {schema_dict}")
    if "project" not in schema_dict[PROP_KEY]:
        _LOGGER.debug("No project section found in schema")

    if SAMPLES_KEY in schema_dict[PROP_KEY]:
        if (
            "items" in schema_dict[PROP_KEY][SAMPLES_KEY]
            and PROP_KEY in schema_dict[PROP_KEY][SAMPLES_KEY]["items"]
        ):
            s_props = schema_dict[PROP_KEY][SAMPLES_KEY]["items"][PROP_KEY]
            for prop, val in s_props.items():
                if "type" in val and val["type"] in ["string", "number", "boolean"]:
                    s_props[prop] = {}
                    s_props[prop]["anyOf"] = [val, {"type": "array", "items": val}]
    else:
        _LOGGER.debug("No samples section found in schema")
    _LOGGER.debug(f"schema processed: {schema_dict}")
    return schema_dict


def read_schema(schema: str | dict) -> list[dict]:
    """
    Safely read schema from YAML-formatted file.

    If the schema imports any other schemas, they will be read recursively.

    Args:
        schema: Path to the schema file or schema in a dict form

    Returns:
        Read schemas

    Raises:
        TypeError: If the schema arg is neither a Mapping nor a file path or
            if the 'imports' sections in any of the schemas is not a list
    """

    def _recursively_read_schemas(
        x: dict, lst: list[dict], parent_folder: str | None
    ) -> list[dict]:
        if "imports" in x:
            if isinstance(x["imports"], list):
                for sch in x["imports"]:
                    if (not is_url(sch)) and (not os.path.isabs(sch)):
                        # resolve relative path
                        if parent_folder is not None:
                            sch = os.path.normpath(os.path.join(parent_folder, sch))
                        else:
                            _LOGGER.warning(
                                f"The schema contains relative path without known parent folder: {sch}"
                            )
                    lst.extend(read_schema(sch))
            else:
                raise TypeError("In schema the 'imports' section has to be a list")
        lst.append(x)
        return lst

    schema_list = []
    schema_folder = None
    if isinstance(schema, str):
        _LOGGER.debug(f"Reading schema: {schema}")
        schema_folder = os.path.split(schema)[0]
        schema = load_yaml(schema)
    if not isinstance(schema, dict):
        raise TypeError(
            f"schema has to be a dict, path to an existing file or URL to a remote one. "
            f"Got: {type(schema)}"
        )
    return _recursively_read_schemas(schema, schema_list, schema_folder)



import os
from collections.abc import Mapping
from copy import deepcopy as dpcpy
from logging import getLogger
from warnings import warn

import pandas as pd
from jsonschema import Draft7Validator
from pandas.core.common import flatten

from ..project import Project
from ..sample import Sample
from ..utils import load_yaml
from .const import PROP_KEY, SAMPLES_KEY, SIZING_KEY, TANGIBLE_KEY
from .exceptions import EidoValidationError, PathAttrNotFoundError
from .schema import preprocess_schema, read_schema

_LOGGER = getLogger(__name__)


def _validate_object(
    obj: Mapping,
    schema: str | dict,
    sample_name_colname: str | bool = False,
) -> None:
    """
    Generic function to validate object against a schema.

    Args:
        obj: An object to validate
        schema: Schema dict to validate against or a path to one
        sample_name_colname: Column name for sample names in error reporting

    Raises:
        EidoValidationError: If validation is unsuccessful
    """
    validator = Draft7Validator(schema)
    _LOGGER.debug(f"{obj},\n {schema}")
    if not validator.is_valid(obj):
        errors = sorted(validator.iter_errors(obj), key=lambda e: e.path)
        errors_by_type = {}

        # Accumulate and restructure error objects by error type
        for error in errors:
            if error.message not in errors_by_type:
                errors_by_type[error.message] = []

            try:
                instance_name = error.instance[sample_name_colname]
            except KeyError:
                instance_name = "project"
            except TypeError:
                instance_name = obj["samples"][error.absolute_path[1]][
                    sample_name_colname
                ]
            errors_by_type[error.message].append(
                {
                    "type": error.message,
                    "message": f"{error.message} on instance {instance_name}",
                    "sample_name": instance_name,
                }
            )

        raise EidoValidationError("Validation failed", errors_by_type)
    else:
        _LOGGER.debug("Validation was successful...")


def validate_project(project: Project, schema: str | dict) -> None:
    """
    Validate a project object against a schema.

    Args:
        project: A project object to validate
        schema: Schema dict to validate against or a path to one

    Raises:
        EidoValidationError: If validation is unsuccessful
    """
    sample_name_colname = project.sample_name_colname
    schema_dicts = read_schema(schema=schema)
    for schema_dict in schema_dicts:
        project_dict = project.to_dict()
        _validate_object(
            project_dict, preprocess_schema(schema_dict), sample_name_colname
        )
        _LOGGER.debug("Project validation successful")


def _validate_sample_object(sample: Sample, schemas: list[dict]) -> None:
    """
    Validate a peppy.Sample object without requiring a reference to peppy.Project.

    Args:
        sample: A sample object to validate
        schemas: List of schemas to validate against or a path to one
    """
    for schema_dict in schemas:
        schema_dict = preprocess_schema(schema_dict)
        sample_schema_dict = schema_dict[PROP_KEY][SAMPLES_KEY]["items"]
        _validate_object(sample.to_dict(), sample_schema_dict)
        _LOGGER.debug(
            f"{getattr(sample, 'sample_name', '')} sample validation successful"
        )


def validate_sample(
    project: Project, sample_name: str | int, schema: str | dict
) -> None:
    """
    Validate the selected sample object against a schema.

    Args:
        project: A project object to validate
        sample_name: Name or index of the sample to validate
        schema: Schema dict to validate against or a path to one

    Raises:
        EidoValidationError: If validation is unsuccessful
    """
    sample = (
        project.samples[sample_name]
        if isinstance(sample_name, int)
        else project.get_sample(sample_name)
    )
    _validate_sample_object(
        sample=sample,
        schemas=read_schema(schema=schema),
    )


def validate_config(project: Project | dict | str, schema: str | dict) -> None:
    """
    Validate the config part of the Project object against a schema.

    Args:
        project: A project object, dict, or path to config file to validate
        schema: Schema dict to validate against or a path to one
    """
    schema_dicts = read_schema(schema=schema)
    for schema_dict in schema_dicts:
        schema_cpy = preprocess_schema(dpcpy(schema_dict))
        try:
            del schema_cpy[PROP_KEY][SAMPLES_KEY]
        except KeyError:
            # Schema doesn't have samples key, which is fine for config-only validation
            pass
        if "required" in schema_cpy:
            try:
                schema_cpy["required"].remove(SAMPLES_KEY)
            except ValueError:
                # SAMPLES_KEY is not in required list, no action needed
                pass
        if isinstance(project, dict):
            _validate_object({"project": project}, schema_cpy)

        elif isinstance(project, str):
            try:
                project_dict = load_yaml(project)
            except (FileNotFoundError, IOError, OSError) as e:
                raise ValueError(
                    f"Please provide a valid yaml config of PEP project; invalid config path: {project}"
                ) from e
            _validate_object({"project": project_dict}, schema_cpy)
        else:
            project_dict = project.to_dict()
            _validate_object(project_dict, schema_cpy)
            _LOGGER.debug("Config validation successful")


def _get_attr_values(obj: Mapping, attrlist: str | list[str]) -> list[str] | None:
    """
    Get value corresponding to each given attribute.

    Args:
        obj: An object to get the attributes from
        attrlist: Names of attributes to retrieve values for

    Returns:
        Value corresponding to each named attribute; None if this Sample's
        value for the attribute is empty/null, or if this Sample lacks the
        indicated attribute
    """
    # If attribute is None, then value is also None.
    if not attrlist:
        return None
    if not isinstance(attrlist, list):
        attrlist = [attrlist]
    # Strings contained here are appended later so shouldn't be null.
    return list(flatten([getattr(obj, attr, "") for attr in attrlist]))


def validate_input_files(
    project: Project,
    schemas: str | dict,
    sample_name: str | int | None = None,
) -> None:
    """
    Determine which of the required and optional files are missing.

    The names of the attributes that are required and/or deemed as inputs
    are sourced from the schema, more specifically from `required_files`
    and `files` sections in samples section:

    - If any of the required files are missing, this function raises an error.
    - If any of the optional files are missing, the function raises a warning.

    Note, this function also performs Sample object validation with jsonschema.

    Args:
        project: Project that defines the samples to validate
        schemas: Schema dict to validate against or a path to one
        sample_name: Name or index of the sample to validate. If None,
            validate all samples in the project

    Raises:
        PathAttrNotFoundError: If any required sample attribute is missing
    """

    if sample_name is None:
        samples = project.samples
    else:
        samples = (
            project.samples[sample_name]
            if isinstance(sample_name, int)
            else project.get_sample(sample_name)
        )
        samples = [samples]

    if isinstance(schemas, str):
        schemas = read_schema(schemas)

    for sample in samples:
        # validate attrs existence first
        _validate_sample_object(schemas=schemas, sample=sample)

        all_inputs = set()
        required_inputs = set()
        schema = schemas[-1]  # use only first schema, in case there are imports
        sample_schema_dict = schema[PROP_KEY][SAMPLES_KEY]["items"]
        if SIZING_KEY in sample_schema_dict:
            all_inputs.update(_get_attr_values(sample, sample_schema_dict[SIZING_KEY]))
        if TANGIBLE_KEY in sample_schema_dict:
            required_inputs = set(
                _get_attr_values(sample, sample_schema_dict[TANGIBLE_KEY])
            )
            all_inputs.update(required_inputs)

        missing_required_inputs = [i for i in required_inputs if not os.path.exists(i)]
        missing_inputs = [i for i in all_inputs if not os.path.exists(i)]
        if missing_inputs:
            warn(
                f"For sample '{getattr(sample, project.sample_table_index)}'. "
                f"Optional inputs not found: {missing_inputs}"
            )
        if missing_required_inputs:
            raise PathAttrNotFoundError(
                f"For sample '{getattr(sample, project.sample_table_index)}'. "
                f"Required inputs not found: {required_inputs}"
            )


def validate_original_samples(samples: str | pd.DataFrame, schema: str | dict) -> None:
    """
    Validate the original samples from the csv table against a schema.

    Args:
        samples: The path to the sample table csv or the dataframe from the table
        schema: Schema dict to validate against or a path to one

    Raises:
        EidoValidationError: If validation is unsuccessful
    """
    if isinstance(samples, str):
        samples = pd.read_csv(samples)

    assist_project = Project.from_pandas(samples)
    for s in assist_project.samples:
        _validate_sample_object(
            sample=s,
            schemas=read_schema(schema=schema),
        )



import pytest

from peppy.pephubclient.pephubclient import PEPHubClient


@pytest.mark.skip(reason="Manual test")
class TestViewsManual:
    def test_get(self):
        ff = PEPHubClient().view.get(
            "databio",
            "bedset1",
            "default",
            "test_view",
        )
        print(ff)

    def test_create(self):
        PEPHubClient().view.create(
            "databio",
            "bedset1",
            "default",
            "test_view",
            sample_list=["orange", "grape1", "apple1"],
        )

    def test_delete(self):
        PEPHubClient().view.delete(
            "databio",
            "bedset1",
            "default",
            "test_view",
        )

    def test_add_sample(self):
        PEPHubClient().view.add_sample(
            "databio",
            "bedset1",
            "default",
            "test_view",
            "name",
        )

    def test_delete_sample(self):
        PEPHubClient().view.remove_sample(
            "databio",
            "bedset1",
            "default",
            "test_view",
            "name",
        )


@pytest.mark.skip(reason="Manual test")
class TestSamplesManual:
    def test_manual(self):
        ff = PEPHubClient().sample.get(
            "databio",
            "bedset1",
            "default",
            "grape1",
        )
        ff

    def test_update(self):
        ff = PEPHubClient().sample.get(
            "databio",
            "bedset1",
            "default",
            "newf",
        )
        ff.update({"shefflab": "test1"})
        ff["sample_type"] = "new_type"
        PEPHubClient().sample.update(
            "databio",
            "bedset1",
            "default",
            "newf",
            sample_dict=ff,
        )

    def test_add(self):
        ff = {
            "genome": "phc_test1",
            "sample_type": "phc_test",
        }
        PEPHubClient().sample.create(
            "databio",
            "bedset1",
            "default",
            "new_2222",
            overwrite=False,
            sample_dict=ff,
        )

    def test_delete(self):
        PEPHubClient().sample.remove(
            "databio",
            "bedset1",
            "default",
            "new_2222",
        )



import sys
from logging import CRITICAL, DEBUG, ERROR, INFO, WARN, Logger

import typer
from logmuse import init_logger

from ..const import PKG_NAME, SAMPLE_NAME_ATTR
from ..project import Project
from .const import CONVERT_CMD, INSPECT_CMD, LOGGING_LEVEL, SUBPARSER_MSGS, VALIDATE_CMD
from .conversion import (
    convert_project,
    get_available_pep_filters,
    pep_conversion_plugins,
)
from .exceptions import EidoFilterError, EidoValidationError
from .inspection import inspect_project
from .validation import validate_config, validate_project, validate_sample

LEVEL_BY_VERBOSITY = [ERROR, CRITICAL, WARN, INFO, DEBUG]

app = typer.Typer()


def _configure_logging(
    verbosity: int | None,
    logging_level: str | None,
    dbg: bool,
) -> str:
    """Mimic old verbosity / logging-level behavior."""
    if dbg:
        level = logging_level or DEBUG
    elif verbosity is not None:
        # Verbosity-framed specification trumps logging_level.
        level = LEVEL_BY_VERBOSITY[verbosity]
    else:
        level = LOGGING_LEVEL
    return level


def _parse_filter_args_str(input: list[str] | None) -> dict[str, str]:
    """
    Parse user input specification.

    Args:
        input: User command line input, formatted as follows:
            [[arg=txt, arg1=txt]]

    Returns:
        Mapping of keys, which are input names and values.
    """
    lst = []
    for i in input or []:
        lst.extend(i)
    return (
        {x.split("=")[0]: x.split("=")[1] for x in lst if "=" in x}
        if lst is not None
        else lst
    )


def print_error_summary(
    errors_by_type: dict[str, list[dict[str, str]]], _LOGGER: Logger
):
    """Print a summary of errors, organized by error type"""
    n_error_types = len(errors_by_type)
    _LOGGER.error(f"Found {n_error_types} types of error:")
    for err_type, items in errors_by_type.items():
        n = len(items)
        msg = f"  - {err_type}: ({n} samples) "
        if n < 50:
            msg += ", ".join(x["sample_name"] for x in items)
        _LOGGER.error(msg)

    if len(errors_by_type) > 1:
        final_msg = f"Validation unsuccessful. {len(errors_by_type)} error types found."
    else:
        final_msg = f"Validation unsuccessful. {len(errors_by_type)} error type found."

    _LOGGER.error(final_msg)


@app.callback()
def common(
    ctx: typer.Context,
    verbosity: int | None = typer.Option(
        None,
        "--verbosity",
        min=0,
        max=len(LEVEL_BY_VERBOSITY) - 1,
        help=f"Choose level of verbosity (default: {None})",
    ),
    logging_level: str | None = typer.Option(
        None,
        "--logging-level",
        help="logging level",
    ),
    dbg: bool = typer.Option(
        False,
        "--dbg",
        help=f"Turn on debug mode (default: {False})",
    ),
):
    ctx.obj = {
        "verbosity": verbosity,
        "logging_level": logging_level,
        "dbg": dbg,
    }

    logger_level = _configure_logging(verbosity, logging_level, dbg)
    logger_kwargs = {"level": logger_level, "devmode": dbg}

    global _LOGGER
    _LOGGER = init_logger(name=PKG_NAME, **logger_kwargs)


@app.command(name=CONVERT_CMD, help=SUBPARSER_MSGS[CONVERT_CMD])
def convert(
    ctx: typer.Context,
    pep: str | None = typer.Argument(
        None,
        metavar="PEP",
        help="Path to a PEP configuration file in yaml format.",
    ),
    st_index: str | None = typer.Option(
        None, "--st-index", help="Sample table index to use"
    ),
    sst_index: str | None = typer.Option(
        None, "--sst-index", help="Subsample table index to use"
    ),
    amendments: list[str] | None = typer.Option(
        None,
        "--amendments",
        help="Names of the amendments to activate.",
    ),
    format_: str = typer.Option(
        "yaml",
        "-f",
        "--format",
        help="Output format (name of filter; use -l to see available).",
    ),
    sample_name: list[str] | None = typer.Option(
        None,
        "-n",
        "--sample-name",
        help="Name of the samples to inspect.",
    ),
    args: list[str] | None = typer.Option(
        None,
        "-a",
        "--args",
        help=("Provide arguments to the filter function (e.g. arg1=val1 arg2=val2)."),
    ),
    list_filters: bool = typer.Option(
        False,
        "-l",
        "--list",
        help="List available filters.",
    ),
    describe: bool = typer.Option(
        False,
        "-d",
        "--describe",
        help="Show description for a given filter.",
    ),
    paths_: list[str] | None = typer.Option(
        None,
        "-p",
        "--paths",
        help="Paths to dump conversion result as key=value pairs.",
    ),
):
    filters = get_available_pep_filters()
    if list_filters:
        _LOGGER.info("Available filters:")
        if len(filters) < 1:
            _LOGGER.info("No available filters")
        for filter_name in filters:
            _LOGGER.info(f" - {filter_name}")
        sys.exit(0)
    if describe:
        if format_ not in filters:
            raise EidoFilterError(
                f"'{format_}' filter not found. Available filters: {', '.join(filters)}"
            )
        filter_functions_by_name = pep_conversion_plugins()
        print(filter_functions_by_name[format_].__doc__)
        sys.exit(0)
    if pep is None:
        typer.echo(ctx.get_help(), err=True)
        _LOGGER.info("The following arguments are required: PEP")
        sys.exit(1)

    if paths_:
        paths = {y[0]: y[1] for y in [x.split("=") for x in paths_]}
    else:
        paths = None

    p = Project(
        pep,
        sample_table_index=st_index,
        subsample_table_index=sst_index,
        amendments=amendments,
    )

    plugin_kwargs = _parse_filter_args_str(args)

    # append paths
    plugin_kwargs["paths"] = paths

    convert_project(p, format_, plugin_kwargs)
    _LOGGER.info("Conversion successful")
    sys.exit(0)


@app.command(name=VALIDATE_CMD, help=SUBPARSER_MSGS[VALIDATE_CMD])
def validate(
    pep: str = typer.Argument(
        None,
        metavar="PEP",
        help="Path to a PEP configuration file in yaml format.",
    ),
    schema: str = typer.Option(
        None,
        "-s",
        "--schema",
        metavar="S",
        help="Path to a PEP schema file in yaml format.",
    ),
    st_index: str | None = typer.Option(
        None,
        "--st-index",
        help=(
            f"Sample table index to use; samples are identified by "
            f"'{SAMPLE_NAME_ATTR}' by default."
        ),
    ),
    sst_index: str | None = typer.Option(
        None,
        "--sst-index",
        help=(
            f"Subsample table index to use; samples are identified by "
            f"'{SAMPLE_NAME_ATTR}' by default."
        ),
    ),
    amendments: list[str] | None = typer.Option(
        None,
        "--amendments",
        help="Names of the amendments to activate.",
    ),
    sample_name: str | None = typer.Option(
        None,
        "-n",
        "--sample-name",
        metavar="S",
        help=(
            "Name or index of the sample to validate. "
            "Only this sample will be validated."
        ),
    ),
    just_config: bool = typer.Option(
        False,
        "-c",
        "--just-config",
        help="Whether samples should be excluded from the validation.",
    ),
):
    if sample_name and just_config:
        raise typer.BadParameter(
            "Use only one of --sample-name or --just-config for 'validate'."
        )
    p = Project(
        pep,
        sample_table_index=st_index,
        subsample_table_index=sst_index,
        amendments=amendments,
    )
    if sample_name:
        try:
            sample_name = int(sample_name)
        except ValueError:
            pass
        _LOGGER.debug(
            f"Comparing Sample ('{pep}') in Project ('{pep}') "
            f"against a schema: {schema}"
        )
        validator = validate_sample
        arguments = [p, sample_name, schema]
    elif just_config:
        _LOGGER.debug(f"Comparing Project ('{pep}') against a schema: {schema}")

        validator = validate_config
        arguments = [p, schema]
    else:
        _LOGGER.debug(f"Comparing Project ('{pep}') against a schema: {schema}")

        validator = validate_project
        arguments = [p, schema]
    try:
        validator(*arguments)
    except EidoValidationError as e:
        print_error_summary(e.errors_by_type, _LOGGER)
        sys.exit(1)
    _LOGGER.info("Validation successful")
    sys.exit(0)


@app.command(name=INSPECT_CMD, help=SUBPARSER_MSGS[INSPECT_CMD])
def inspect(
    pep: str = typer.Argument(
        None,
        metavar="PEP",
        help="Path to a PEP configuration file in yaml format.",
    ),
    st_index: str | None = typer.Option(
        None,
        "--st-index",
        help=(
            f"Sample table index to use; samples are identified by "
            f"'{SAMPLE_NAME_ATTR}' by default."
        ),
    ),
    sst_index: str | None = typer.Option(
        None,
        "--sst-index",
        help=(
            f"Subsample table index to use; samples are identified by "
            f"'{SAMPLE_NAME_ATTR}' by default."
        ),
    ),
    amendments: list[str] | None = typer.Option(
        None,
        "--amendments",
        help="Names of the amendments to activate.",
    ),
    sample_name: list[str] | None = typer.Option(
        None,
        "-n",
        "--sample-name",
        metavar="SN",
        help="Name of the samples to inspect.",
    ),
    attr_limit: int = typer.Option(
        10,
        "-l",
        "--attr-limit",
        help="Number of sample attributes to display.",
    ),
):
    p = Project(
        pep,
        sample_table_index=st_index,
        subsample_table_index=sst_index,
        amendments=amendments,
    )
    inspect_project(p, sample_name, attr_limit)



import logging

# import peppy
from ...project import Project
from ..constants import PEPHUB_VIEW_SAMPLE_URL, PEPHUB_VIEW_URL, ResponseStatusCodes
from ..exceptions import ResponseError
from ..helpers import RequestManager
from ..models import ProjectDict

_LOGGER = logging.getLogger("pephubclient")


class PEPHubView(RequestManager):
    """
    Class for managing views in PEPhub.

    Provides methods for getting, creating, updating and removing views.
    This class aims to wrap the Views API for easier maintenance and
    better user experience.
    """

    def __init__(self, jwt_data: str = None):
        """
        Initialize PEPHubView.

        Args:
            jwt_data: jwt token for authorization
        """

        self.__jwt_data = jwt_data

    def get(
        self, namespace: str, name: str, tag: str, view_name: str, raw: bool = False
    ) -> Project | dict:
        """
        Get view from project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            view_name: name of the view
            raw: if True, return raw response

        Returns:
            peppy.Project object or dictionary of the project (view).
        """
        url = self._build_view_request_url(
            namespace=namespace, name=name, view_name=view_name
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="GET", url=url, headers=self.parse_header(self.__jwt_data)
        )
        if response.status_code == ResponseStatusCodes.OK:
            output = self.decode_response(response, output_json=True)
            if raw:
                return output
            output = ProjectDict(**output).model_dump(by_alias=True)
            return Project.from_dict(output)
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError("View does not exist, or you are unauthorized.")
        else:
            raise ResponseError(
                f"Internal server error. Unexpected return value. Error: {response.status_code}"
            )

    def create(
        self,
        namespace: str,
        name: str,
        tag: str,
        view_name: str,
        description: str = None,
        sample_list: list = None,
        no_fail: bool = False,
    ):
        """
        Create view in project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            description: description of the view
            view_name: name of the view
            sample_list: list of sample names
            no_fail: whether to raise an error if view was not added to the
                project
        """

        if not sample_list or not isinstance(sample_list, list):
            raise ValueError("Sample list must be a list of sample names.")

        url = self._build_view_request_url(
            namespace=namespace, name=name, view_name=view_name
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="POST",
            url=url,
            headers=self.parse_header(self.__jwt_data),
            params={"description": description, "no_fail": no_fail},
            json=sample_list,
        )
        if response.status_code == ResponseStatusCodes.ACCEPTED:
            _LOGGER.info(
                f"View '{view_name}' created in project '{namespace}/{name}:{tag}' successfully."
            )
            return None
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError(
                f"Project '{namespace}/{name}:{tag}' or one of the samples does not exist."
            )
        elif response.status_code == ResponseStatusCodes.CONFLICT:
            raise ResponseError(f"View '{view_name}' already exists in the project.")
        else:
            raise ResponseError(f"Unexpected return value.{response.status_code}")

    def delete(self, namespace: str, name: str, tag: str, view_name: str) -> None:
        """
        Delete view from project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            view_name: name of the view
        """
        url = self._build_view_request_url(
            namespace=namespace, name=name, view_name=view_name
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="DELETE", url=url, headers=self.parse_header(self.__jwt_data)
        )

        if response.status_code == ResponseStatusCodes.ACCEPTED:
            _LOGGER.info(
                f"View '{view_name}' deleted from project '{namespace}/{name}:{tag}' successfully."
            )
            return None
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError("View does not exists, or you are unauthorized.")
        elif response.status_code == ResponseStatusCodes.UNAUTHORIZED:
            raise ResponseError("You are unauthorized to delete this view.")
        else:
            raise ResponseError("Unexpected return value. ")

    def add_sample(
        self,
        namespace: str,
        name: str,
        tag: str,
        view_name: str,
        sample_name: str,
    ):
        """
        Add sample to view in project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            view_name: name of the view
            sample_name: name of the sample
        """
        url = self._build_view_request_url(
            namespace=namespace,
            name=name,
            view_name=view_name,
            sample_name=sample_name,
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="POST",
            url=url,
            headers=self.parse_header(self.__jwt_data),
        )
        if response.status_code == ResponseStatusCodes.ACCEPTED:
            _LOGGER.info(
                f"Sample '{sample_name}' added to view '{view_name}' in project '{namespace}/{name}:{tag}' successfully."
            )
            return None
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError(
                f"Sample '{sample_name}' or project {namespace}/{name}:{tag} does not exist."
            )
        elif response.status_code == ResponseStatusCodes.CONFLICT:
            raise ResponseError(f"Sample '{sample_name}' already exists in the view.")
        else:
            raise ResponseError(
                f"Unexpected return value. Error: {response.status_code}"
            )

    def remove_sample(
        self,
        namespace: str,
        name: str,
        tag: str,
        view_name: str,
        sample_name: str,
    ):
        """
        Remove sample from view in project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            view_name: name of the view
            sample_name: name of the sample
        """
        url = self._build_view_request_url(
            namespace=namespace,
            name=name,
            view_name=view_name,
            sample_name=sample_name,
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="DELETE",
            url=url,
            headers=self.parse_header(self.__jwt_data),
        )
        if response.status_code == ResponseStatusCodes.ACCEPTED:
            _LOGGER.info(
                f"Sample '{sample_name}' removed from view '{view_name}' in project '{namespace}/{name}:{tag}' successfully."
            )
            return None
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError(
                f"Sample '{sample_name}' or project {namespace}/{name}:{tag} does not exist. "
            )
        elif response.status_code == ResponseStatusCodes.UNAUTHORIZED:
            raise ResponseError(
                "You are unauthorized to remove this sample from the view."
            )
        else:
            raise ResponseError(
                f"Unexpected return value. Error: {response.status_code}"
            )

    @staticmethod
    def _build_view_request_url(
        namespace: str, name: str, view_name: str, sample_name: str = None
    ):
        """
        Build URL for view request.

        Args:
            namespace: namespace of project
            name: name of project
            view_name: name of view

        Returns:
            URL.
        """
        if sample_name:
            return PEPHUB_VIEW_SAMPLE_URL.format(
                namespace=namespace,
                project=name,
                view_name=view_name,
                sample_name=sample_name,
            )
        return PEPHUB_VIEW_URL.format(
            namespace=namespace,
            project=name,
            view_name=view_name,
        )



from peppy.eido.conversion import (
    convert_project,
    get_available_pep_filters,
    pep_conversion_plugins,
    run_filter,
)
from peppy.project import Project


class TestConversionInfrastructure:
    def test_plugins_are_read(self):
        avail_filters = get_available_pep_filters()
        assert isinstance(avail_filters, list)

    def test_plugins_contents(self):
        avail_plugins = pep_conversion_plugins()
        avail_filters = get_available_pep_filters()
        assert all(
            [plugin_name in avail_filters for plugin_name in avail_plugins.keys()]
        )

    def test_plugins_are_callable(self):
        avail_plugins = pep_conversion_plugins()
        assert all(
            [callable(plugin_fun) for plugin_name, plugin_fun in avail_plugins.items()]
        )

    def test_basic_filter(self, save_result_mock, project_object):
        conv_result = run_filter(
            project_object,
            "basic",
            verbose=False,
            plugin_kwargs={"paths": {"project": "out/basic_prj.txt"}},
        )

        assert save_result_mock.called
        assert conv_result["project"] == str(project_object)

    def test_csv_filter(
        self, save_result_mock, taxprofiler_project, taxprofiler_csv_multiline_output
    ):
        conv_result = run_filter(
            taxprofiler_project,
            "csv",
            verbose=False,
            plugin_kwargs={"paths": {"samples": "out/basic_prj.txt"}},
        )

        assert save_result_mock.called
        assert conv_result["samples"] == taxprofiler_csv_multiline_output

    def test_csv_filter_handles_empty_fasta_correctly(
        self,
        project_pep_with_fasta_column,
        output_pep_with_fasta_column,
        save_result_mock,
    ):
        conv_result = run_filter(
            project_pep_with_fasta_column,
            "csv",
            verbose=False,
            plugin_kwargs={"paths": {"samples": "out/basic_prj.txt"}},
        )

        assert save_result_mock.called
        assert conv_result == {"samples": output_pep_with_fasta_column}

    def test_eido_csv_filter_filters_nextflow_taxprofiler_input_correctly(
        self,
        project_pep_nextflow_taxprofiler,
        output_pep_nextflow_taxprofiler,
        save_result_mock,
    ):
        conv_result = run_filter(
            project_pep_nextflow_taxprofiler,
            "csv",
            verbose=False,
            plugin_kwargs={"paths": {"samples": "out/basic_prj.txt"}},
        )

        assert save_result_mock.called
        assert conv_result == {"samples": output_pep_nextflow_taxprofiler}

    def test_multiple_subsamples(self, test_multiple_subs):
        project = Project(test_multiple_subs, sample_table_index="sample_id")

        conversion = convert_project(
            project,
            "csv",
        )
        assert isinstance(conversion["samples"], str)
        conversion = convert_project(
            project,
            "basic",
        )
        assert isinstance(conversion["project"], str)
        conversion = convert_project(
            project,
            "yaml",
        )
        assert isinstance(conversion["project"], str)
        conversion = convert_project(
            project,
            "yaml-samples",
        )
        assert isinstance(conversion["samples"], str)



import os
import zipfile
from contextlib import suppress
from pathlib import Path

import pandas
import yaml

from .exceptions import PEPExistsError


class FilesManager:
    @staticmethod
    def save_jwt_data_to_file(path: str, jwt_data: str) -> None:
        """Save jwt to provided path."""
        Path(os.path.dirname(path)).mkdir(parents=True, exist_ok=True)
        with open(path, "w") as f:
            f.write(jwt_data)

    @staticmethod
    def load_jwt_data_from_file(path: str) -> str:
        """Open the file with username and ID and load this data."""
        with suppress(FileNotFoundError):
            with open(path, "r") as f:
                return f.read()

    @staticmethod
    def create_project_folder(
        parent_path: str,
        folder_name: str,
    ) -> str:
        """
        Create new project folder.

        Args:
            parent_path: parent path to create folder in
            folder_name: folder name

        Returns:
            folder_path.
        """
        if parent_path:
            if not Path(parent_path).exists():
                raise OSError(
                    f"Parent path does not exist. Provided path: {parent_path}"
                )
        folder_path = os.path.join(parent_path or os.getcwd(), folder_name)
        Path(folder_path).mkdir(parents=True, exist_ok=True)
        return folder_path

    @staticmethod
    def save_yaml(config: dict, full_path: str, not_force: bool = False):
        FilesManager.check_writable(path=full_path, force=not not_force)
        with open(full_path, "w") as outfile:
            yaml.dump(config, outfile, default_flow_style=False)

    @staticmethod
    def save_pandas(df: pandas.DataFrame, full_path: str, not_force: bool = False):
        FilesManager.check_writable(path=full_path, force=not not_force)
        df.to_csv(full_path, index=False)

    @staticmethod
    def file_exists(full_path: str) -> bool:
        return os.path.isfile(full_path)

    @staticmethod
    def delete_file_if_exists(filename: str) -> None:
        with suppress(FileNotFoundError):
            os.remove(filename)
            print(
                f"\033[38;5;11m{f'File was deleted successfully -> {filename}'}\033[0m"
            )

    @staticmethod
    def check_writable(path: str, force: bool = True):
        if not force and os.path.isfile(path):
            raise PEPExistsError(f"File already exists and won't be updated: {path}")

    @staticmethod
    def save_zip_file(files_dict: dict, file_path: str, force: bool = False) -> None:
        """
        Save zip file with provided files as dict.

        Args:
            files_dict: dict with files to save. e.g. {"file1.txt": "file1 content"}
            file_path: filename to save zip file to
            force: overwrite file if exists
        """
        FilesManager.check_writable(path=file_path, force=force)
        with zipfile.ZipFile(
            file_path, mode="w", compression=zipfile.ZIP_DEFLATED
        ) as zf:
            for name, res in files_dict.items():
                zf.writestr(name, str.encode(res))



import os
from collections.abc import Iterable
from logging import getLogger
from warnings import catch_warnings

from ubiquerg import size

from .. import Project, Sample
from .const import (
    ALL_INPUTS_KEY,
    INPUT_FILE_SIZE_KEY,
    MISSING_KEY,
    PROP_KEY,
    REQUIRED_INPUTS_KEY,
    SAMPLES_KEY,
    SIZING_KEY,
    TANGIBLE_KEY,
)
from .schema import read_schema
from .validation import _get_attr_values, _validate_sample_object

_LOGGER = getLogger(__name__)


def inspect_project(
    p: Project, sample_names: list[str] | None = None, max_attr: int = 10
) -> None:
    """
    Print inspection info: Project or, if sample_names argument is provided, matched samples.

    Args:
        p: Project to inspect
        sample_names: List of samples to inspect
        max_attr: Max number of sample attributes to display
    """
    if sample_names:
        samples = p.get_samples(sample_names)
        if not samples:
            print("No samples matched by names: {}".format(sample_names))
            return
        for s in samples:
            print(s.__str__(max_attr=max_attr))
            print("\n")
        return
    print(p)
    return


def get_input_files_size(
    sample: Sample, schema: str | list[dict]
) -> dict[str, list[str] | set[str] | float]:
    """
    Determine which of this Sample's required attributes/files are missing and calculate sizes.

    The names of the attributes that are required and/or deemed as inputs
    are sourced from the schema, more specifically from required_input_attrs
    and input_attrs sections in samples section. Note, this function does
    perform actual Sample object validation with jsonschema.

    Args:
        sample: Sample to investigate
        schema: Schema dict to validate against or a path to one

    Returns:
        Dictionary with validation data, i.e missing, required_inputs,
        all_inputs, input_file_size

    Raises:
        ValidationError: If any required sample attribute is missing
    """

    def _compute_input_file_size(inputs: Iterable[str]) -> float:
        """Compute total size of input files."""
        with catch_warnings(record=True) as w:
            total_bytes = sum(
                size(f, size_str=False) or 0.0
                for f in inputs
                if f != "" and f is not None
            )
            if w:
                _LOGGER.warning(
                    f"{len(w)} input files missing, job input size was "
                    f"not calculated accurately"
                )
        return total_bytes / (1024**3)

    if isinstance(schema, str):
        schema = read_schema(schema)

    # first, validate attrs existence using jsonschema
    _validate_sample_object(schemas=schema, sample=sample)

    all_inputs = set()
    required_inputs = set()
    schema = schema[-1]  # use only first schema, in case there are imports
    sample_schema_dict = schema[PROP_KEY][SAMPLES_KEY]["items"]
    if SIZING_KEY in sample_schema_dict:
        all_inputs.update(_get_attr_values(sample, sample_schema_dict[SIZING_KEY]))
    if TANGIBLE_KEY in sample_schema_dict:
        required_inputs = set(
            _get_attr_values(sample, sample_schema_dict[TANGIBLE_KEY])
        )
        all_inputs.update(required_inputs)
    with catch_warnings(record=True) as w:
        # input_file_size = sum(
        #     [
        #         size(f, size_str=False) or 0.0
        #         for f in all_inputs
        #         if f != "" and f is not None
        #     ]
        # ) / (1024**3)
        input_file_size = _compute_input_file_size(all_inputs)
        if w:
            _LOGGER.warning(
                f"{len(w)} input files missing, job input size was "
                f"not calculated accurately"
            )

    return {
        MISSING_KEY: [i for i in required_inputs if not os.path.exists(i)],
        REQUIRED_INPUTS_KEY: required_inputs,
        ALL_INPUTS_KEY: all_inputs,
        INPUT_FILE_SIZE_KEY: input_file_size,
    }



from typing import Literal

import urllib3
from pydantic import ValidationError
from typing_extensions import deprecated
from ubiquerg import parse_registry_path

from ..const import CONFIG_KEY, NAME_KEY
from ..project import Project
from .constants import (
    PATH_TO_FILE_WITH_JWT,
    PEPHUB_PEP_API_BASE_URL,
    PEPHUB_PEP_SEARCH_URL,
    PEPHUB_PUSH_URL,
    RegistryPath,
    ResponseStatusCodes,
)
from .exceptions import IncorrectQueryStringError, ResponseError
from .files_manager import FilesManager
from .helpers import MessageHandler, RequestManager, save_pep
from .models import (
    ProjectAnnotationModel,
    ProjectDict,
    ProjectUploadData,
    SearchReturnModel,
)
from .modules.sample import PEPHubSample
from .modules.view import PEPHubView
from .pephub_oauth.pephub_oauth import PEPHubAuth

urllib3.disable_warnings()


class PEPHubClient(RequestManager):
    def __init__(self):
        self.__jwt_data = FilesManager.load_jwt_data_from_file(PATH_TO_FILE_WITH_JWT)

        self.__view = PEPHubView(self.__jwt_data)
        self.__sample = PEPHubSample(self.__jwt_data)

    @property
    def view(self) -> PEPHubView:
        return self.__view

    @property
    def sample(self) -> PEPHubSample:
        return self.__sample

    def login(self) -> None:
        """Log in to PEPhub."""
        user_token = PEPHubAuth().login_to_pephub()

        FilesManager.save_jwt_data_to_file(PATH_TO_FILE_WITH_JWT, user_token)
        self.__jwt_data = FilesManager.load_jwt_data_from_file(PATH_TO_FILE_WITH_JWT)

    def logout(self) -> None:
        """Log out from PEPhub."""
        FilesManager.delete_file_if_exists(PATH_TO_FILE_WITH_JWT)
        self.__jwt_data = None

    def pull(
        self,
        project_registry_path: str,
        force: bool | None = False,
        zip: bool | None = False,
        output: str | None = None,
    ) -> None:
        """
        Download project locally.

        Args:
            project_registry_path: Project registry path in PEPhub
                (e.g. databio/base:default)
            force: if project exists, overwrite it.
            zip: if True, save project as zip file
            output: path where project will be saved
        """
        project_dict = self.load_raw_pep(
            registry_path=project_registry_path,
        )

        save_pep(
            project=project_dict,
            reg_path=project_registry_path,
            force=force,
            project_path=output,
            zip=zip,
        )

    def load_project(
        self,
        project_registry_path: str,
        query_param: dict | None = None,
    ) -> Project:
        """
        Load peppy project from PEPhub in Project object.

        Args:
            project_registry_path: registry path of the project
            query_param: query parameters used in get request

        Returns:
            peppy project.
        """
        raw_pep = self.load_raw_pep(project_registry_path, query_param)
        peppy_project = Project().from_dict(raw_pep)
        return peppy_project

    def push(
        self,
        cfg: str,
        namespace: str,
        name: str | None = None,
        tag: str | None = None,
        is_private: bool | None = False,
        force: bool | None = False,
    ) -> None:
        """
        Push (upload/update) project to Pephub using config/csv path.

        Args:
            cfg: Project config file (YAML) or sample table (CSV/TSV)
                with one row per sample to constitute project
            namespace: namespace
            name: project name
            tag: project tag
            is_private: Specifies whether project should be private [Default= False]
            force: Force push to the database. Use it to update, or upload
                project. [Default= False]
        """
        peppy_project = Project(cfg=cfg)
        self.upload(
            project=peppy_project,
            namespace=namespace,
            name=name,
            tag=tag,
            is_private=is_private,
            force=force,
        )

    def upload(
        self,
        project: Project,
        namespace: str,
        name: str = None,
        tag: str = None,
        is_private: bool = False,
        force: bool = True,
    ) -> None:
        """
        Upload peppy project to the PEPhub.

        Args:
            project: Project object that has to be uploaded to the DB
            namespace: namespace
            name: project name
            tag: project tag
            is_private: Make project private
            force: overwrite project if it exists, use it to update, or upload
                project.
        """
        pep_dict = project.to_dict(
            extended=True,
            orient="records",
        )
        if name:
            pep_dict[CONFIG_KEY][NAME_KEY] = name

        upload_data = ProjectUploadData(
            pep_dict=pep_dict,
            tag=tag,
            is_private=is_private,
            overwrite=force,
        )
        pephub_response = self.send_request(
            method="POST",
            url=self._build_push_request_url(namespace=namespace),
            headers=self.parse_header(self.__jwt_data),
            json=upload_data.model_dump(),
            cookies=None,
        )
        if pephub_response.status_code == ResponseStatusCodes.ACCEPTED:
            MessageHandler.print_success(
                f"Project '{namespace}/{name}:{upload_data.tag}' was successfully uploaded"
            )
        elif pephub_response.status_code == ResponseStatusCodes.CONFLICT:
            raise ResponseError(
                "Project already exists. Set force to overwrite project."
            )
        elif pephub_response.status_code == ResponseStatusCodes.UNAUTHORIZED:
            raise ResponseError("Unauthorized! Failure in uploading project.")
        elif pephub_response.status_code == ResponseStatusCodes.FORBIDDEN:
            raise ResponseError(
                "User does not have permission to write to this namespace!"
            )
        else:
            detail = ""
            try:
                detail = self.decode_response(pephub_response, output_json=True).get(
                    "detail", ""
                )
            except Exception:
                pass
            raise ResponseError(
                f"Unexpected Response Error. {pephub_response.status_code}: {detail}"
                if detail
                else f"Unexpected Response Error. {pephub_response.status_code}"
            )
        return None

    def find_project(
        self,
        namespace: str,
        query_string: str = "",
        limit: int = 100,
        offset: int = 0,
        filter_by: Literal["submission_date", "last_update_date"] = None,
        start_date: str = None,
        end_date: str = None,
    ) -> SearchReturnModel:
        """
        Find project in specific namespace and return list of PEP annotation.

        Args:
            namespace: Namespace where to search for projects
            query_string: Search query
            limit: Return limit
            offset: Return offset
            filter_by: Use filter date. Option: [submission_date, last_update_date]
            start_date: filter beginning date
            end_date: filter end date (if none today's date is used)
        """

        query_param = {
            "q": query_string,
            "limit": limit,
            "offset": offset,
        }
        if filter_by in ["submission_date", "last_update_date"]:
            query_param["filter_by"] = filter_by
            query_param["filter_start_date"] = start_date
            if end_date:
                query_param["filter_end_date"] = end_date

        url = self._build_project_search_url(
            namespace=namespace,
            query_param=query_param,
        )

        pephub_response = self.send_request(
            method="GET",
            url=url,
            headers=self.parse_header(self.__jwt_data),
            json=None,
            cookies=None,
        )
        if pephub_response.status_code == ResponseStatusCodes.OK:
            decoded_response = self.decode_response(pephub_response, output_json=True)
            project_list = []
            for project_found in decoded_response["results"]:
                project_list.append(ProjectAnnotationModel(**project_found))
            return SearchReturnModel(**decoded_response)

    @deprecated("This method is deprecated. Use load_raw_pep instead.")
    def _load_raw_pep(
        self,
        registry_path: str,
        jwt_data: str | None = None,
        query_param: dict | None = None,
    ) -> dict:
        """
        This method is deprecated. Use load_raw_pep instead.

        Request PEPhub and return the requested project as Project object.

        Args:
            registry_path: Project namespace, eg. "geo/GSE124224:tag"
            query_param: Optional variables to be passed to PEPhub

        Returns:
            Raw project in dict.
        """
        return self.load_raw_pep(registry_path, query_param)

    def load_raw_pep(
        self,
        registry_path: str,
        query_param: dict | None = None,
    ) -> dict:
        """
        Request PEPhub and return the requested project as Project object.

        Args:
            registry_path: Project namespace, eg. "geo/GSE124224:tag"
            query_param: Optional variables to be passed to PEPhub

        Returns:
            Raw project in dict.
        """
        query_param = query_param or {}
        query_param["raw"] = "true"

        self._set_registry_data(registry_path)
        pephub_response = self.send_request(
            method="GET",
            url=self._build_pull_request_url(query_param=query_param),
            headers=self.parse_header(self.__jwt_data),
            cookies=None,
        )
        if pephub_response.status_code == ResponseStatusCodes.OK:
            decoded_response = self.decode_response(pephub_response, output_json=True)
            correct_proj_dict = ProjectDict(**decoded_response)

            # This step is necessary because of this issue: https://github.com/pepkit/pephub/issues/124
            return correct_proj_dict.model_dump(by_alias=True)

        if pephub_response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError("File does not exist, or you are unauthorized.")
        if pephub_response.status_code == ResponseStatusCodes.INTERNAL_ERROR:
            raise ResponseError(
                f"Internal server error. Unexpected return value. Error: {pephub_response.status_code}"
            )

    def _set_registry_data(self, query_string: str) -> None:
        """
        Parse provided query string to extract project name, sample name, etc.

        Args:
            query_string: Passed by user. Contain information needed to locate
                the project.
        """
        try:
            self.registry_path = RegistryPath(**parse_registry_path(query_string))
        except (ValidationError, TypeError):
            raise IncorrectQueryStringError(query_string=query_string)

    def _build_pull_request_url(self, query_param: dict = None) -> str:
        """
        Build request for getting projects from pephub.

        Args:
            query_param: dict of parameters used in query string

        Returns:
            url string.
        """
        query_param = query_param or {}
        query_param["tag"] = self.registry_path.tag

        endpoint = self.registry_path.namespace + "/" + self.registry_path.item

        variables_string = self.parse_query_param(query_param)
        endpoint += variables_string

        return PEPHUB_PEP_API_BASE_URL + endpoint

    @staticmethod
    def _build_project_search_url(namespace: str, query_param: dict = None) -> str:
        """
        Build request for searching projects from pephub.

        Args:
            query_param: dict of parameters used in query string

        Returns:
            url string.
        """

        variables_string = RequestManager.parse_query_param(query_param)
        endpoint = variables_string

        return PEPHUB_PEP_SEARCH_URL.format(namespace=namespace) + endpoint

    @staticmethod
    def _build_push_request_url(namespace: str) -> str:
        """
        Build project upload request used in pephub.

        Args:
            namespace: namespace where project will be uploaded

        Returns:
            url string.
        """
        return PEPHUB_PUSH_URL.format(namespace=namespace)



from importlib.metadata import version

import typer

from .const import PKG_NAME
from .eido.cli import app as eido_app
from .pephubclient.cli import app as phc_app


def version_callback(value: bool):
    if value:
        typer.echo(f"{PKG_NAME} version: {version(PKG_NAME)}")
        raise typer.Exit()


app = typer.Typer(help=f"{PKG_NAME} - Portable Encapsulated Projects toolkit")


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", "-v", callback=version_callback, help="package version"
    ),
):
    pass


app.add_typer(phc_app, name="phc", help="Client for the PEPhub server")
app.add_typer(eido_app, name="eido", help="PEP validation, conversion, and inspection")


def main():
    app(prog_name=PKG_NAME)



import typer

from .helpers import call_client_func
from .pephubclient import PEPHubClient

_client = PEPHubClient()

app = typer.Typer()


@app.command()
def login():
    """
    Login to PEPhub
    """
    call_client_func(_client.login)


@app.command()
def logout():
    """
    Logout
    """
    _client.logout()


@app.command()
def pull(
    project_registry_path: str,
    force: bool = typer.Option(False, help="Overwrite project if it exists."),
    zip: bool = typer.Option(False, help="Save project as zip file."),
    output: str = typer.Option(None, help="Output directory."),
):
    """
    Download and save project locally.
    """
    call_client_func(
        _client.pull,
        project_registry_path=project_registry_path,
        force=force,
        output=output,
        zip=zip,
    )


@app.command()
def push(
    cfg: str = typer.Argument(
        ...,
        help="Project config file (YAML) or sample table (CSV/TSV)"
        "with one row per sample to constitute project",
    ),
    namespace: str = typer.Option(..., help="Project namespace"),
    name: str = typer.Option(..., help="Project name"),
    tag: str = typer.Option(None, help="Project tag"),
    force: bool = typer.Option(
        False, help="Force push to the database. Use it to update, or upload project."
    ),
    is_private: bool = typer.Option(False, help="Upload project as private."),
):
    """
    Upload/update project in PEPhub
    """

    call_client_func(
        _client.push,
        cfg=cfg,
        namespace=namespace,
        name=name,
        tag=tag,
        is_private=is_private,
        force=force,
    )



import urllib

import pytest

from peppy import Project
from peppy.eido.exceptions import EidoValidationError, PathAttrNotFoundError
from peppy.eido.validation import (
    validate_config,
    validate_input_files,
    validate_original_samples,
    validate_project,
    validate_sample,
)
from peppy.utils import load_yaml


def _check_remote_file_accessible(url):
    try:
        code = urllib.request.urlopen(url).getcode()
    except (urllib.error.URLError, OSError):
        pytest.skip(f"Remote file not found: {url}")
    else:
        if code != 200:
            pytest.skip(f"Return code: {code}. Remote file not found: {url}")


class TestProjectValidation:
    def test_validate_works(self, project_object, schema_file_path):
        validate_project(project=project_object, schema=schema_file_path)

    def test_validate_detects_invalid(self, project_object, schema_invalid_file_path):
        with pytest.raises(EidoValidationError):
            validate_project(project=project_object, schema=schema_invalid_file_path)

    def test_validate_detects_invalid_imports(
        self, project_object, schema_imports_file_path
    ):
        with pytest.raises(EidoValidationError):
            validate_project(project=project_object, schema=schema_imports_file_path)

    def test_validate_imports_with_rel_path(
        self, path_pep_for_schema_with_rel_path, schema_rel_path_imports_file_path
    ):
        pep_project = Project(path_pep_for_schema_with_rel_path)
        validate_project(project=pep_project, schema=schema_rel_path_imports_file_path)

    def test_validate_converts_samples_to_private_attr(
        self, project_object, schema_samples_file_path
    ):
        """
        In peppy.Project the list of peppy.Sample objects is
        accessible via _samples attr.
        To make the schema creation more accessible for eido users
        samples->_samples key conversion has been implemented
        """
        validate_project(project=project_object, schema=schema_samples_file_path)

    def test_validate_works_with_dict_schema(self, project_object, schema_file_path):
        validate_project(project=project_object, schema=load_yaml(schema_file_path))

    @pytest.mark.parametrize("schema_arg", [1, None, [1, 2, 3]])
    def test_validate_raises_error_for_incorrect_schema_type(
        self, project_object, schema_arg
    ):
        with pytest.raises(TypeError):
            validate_project(project=project_object, schema=schema_arg)


class TestSampleValidation:
    @pytest.mark.parametrize("sample_name", [0, 1, "GSM1558746"])
    def test_validate_works(
        self, project_object, sample_name, schema_samples_file_path
    ):
        validate_sample(
            project=project_object,
            sample_name=sample_name,
            schema=schema_samples_file_path,
        )

    @pytest.mark.parametrize("sample_name", [22, "bogus_sample_name"])
    def test_validate_raises_error_for_incorrect_sample_name(
        self, project_object, sample_name, schema_samples_file_path
    ):
        with pytest.raises((ValueError, IndexError)):
            validate_sample(
                project=project_object,
                sample_name=sample_name,
                schema=schema_samples_file_path,
            )

    @pytest.mark.parametrize("sample_name", [0, 1, "GSM1558746"])
    def test_validate_detects_invalid(
        self, project_object, sample_name, schema_sample_invalid_file_path
    ):
        with pytest.raises(EidoValidationError):
            validate_sample(
                project=project_object,
                sample_name=sample_name,
                schema=schema_sample_invalid_file_path,
            )

    def test_original_sample(self, project_table_path, schema_samples_file_path):
        validate_original_samples(project_table_path, schema_samples_file_path)


class TestConfigValidation:
    def test_validate_succeeds_on_invalid_sample(
        self, project_object, schema_sample_invalid_file_path
    ):
        validate_config(project=project_object, schema=schema_sample_invalid_file_path)

    def test_validate_on_yaml_dict(
        self, project_file_path, schema_sample_invalid_file_path
    ):
        validate_config(
            project=project_file_path, schema=schema_sample_invalid_file_path
        )


class TestRemoteValidation:
    @pytest.mark.parametrize("schema_url", ["http://schema.databio.org/pep/2.0.0.yaml"])
    def test_validate_works_with_remote_schemas(self, project_object, schema_url):
        _check_remote_file_accessible(schema_url)
        validate_project(project=project_object, schema=schema_url)
        validate_config(project=project_object, schema=schema_url)
        validate_sample(project=project_object, schema=schema_url, sample_name=0)


class TestImportsValidation:
    def test_validate(self, project_object, schema_file_path):
        validate_project(project=project_object, schema=schema_file_path)


class TestProjectWithoutConfigValidation:
    @pytest.mark.parametrize(
        "remote_pep_cfg",
        [
            "https://raw.githubusercontent.com/pepkit/example_peps/master/example_basic/sample_table.csv"
        ],
    )
    def test_validate_works(self, schema_file_path, remote_pep_cfg):
        _check_remote_file_accessible(remote_pep_cfg)
        validate_project(
            project=Project(
                remote_pep_cfg
            ),  # create Project object from a remote sample table
            schema=schema_file_path,
        )

    @pytest.mark.parametrize(
        "remote_pep_cfg",
        [
            "https://raw.githubusercontent.com/pepkit/example_peps/master/example_basic/sample_table.csv"
        ],
    )
    def test_validate_detects_invalid(self, schema_invalid_file_path, remote_pep_cfg):
        _check_remote_file_accessible(remote_pep_cfg)
        with pytest.raises(EidoValidationError):
            validate_project(
                project=Project(remote_pep_cfg), schema=schema_invalid_file_path
            )

    def test_validate_file_existence(
        self, test_file_existing_pep, test_file_existing_schema
    ):
        schema_path = test_file_existing_schema
        prj = Project(test_file_existing_pep)
        with pytest.raises(PathAttrNotFoundError):
            validate_input_files(prj, schema_path)

    def test_validation_values(self, test_schema_value_check, test_file_value_check):
        schema_path = test_schema_value_check
        prj = Project(test_file_value_check)
        with pytest.raises(EidoValidationError):
            validate_project(project=prj, schema=schema_path)



"""Built-in PEP filters."""

from .output_formatters import MultilineOutputFormatter


def basic_pep_filter(p, **kwargs) -> dict[str, str]:
    """
    Basic PEP filter, that does not convert the Project object.

    This filter can save the PEP representation to file, if kwargs include `path`.

    Args:
        p: A Project to run filter on.
    """
    return {"project": str(p)}


def yaml_samples_pep_filter(p, **kwargs) -> dict[str, str]:
    """
    YAML samples PEP filter, that returns only Sample object representations.

    This filter can save the YAML to file, if kwargs include `path`.

    Args:
        p: A Project to run filter on.
    """
    from yaml import dump

    samples_yaml = []
    for s in p.samples:
        samples_yaml.append(s.to_dict())

    return {"samples": dump(samples_yaml, default_flow_style=False)}


def yaml_pep_filter(p, **kwargs) -> dict[str, str]:
    """
    YAML PEP filter, that returns Project object representation.

    This filter can save the YAML to file, if kwargs include `path`.

    Args:
        p: A Project to run filter on.
    """
    from yaml import dump

    return {"project": dump(p.config, default_flow_style=False)}


def csv_pep_filter(p, **kwargs) -> dict[str, str]:
    """
    CSV PEP filter, that returns Sample object representations.

    This filter can save the CSVs to files, if kwargs include
    `sample_table_path` and/or `subsample_table_path`.

    Args:
        p: A Project to run filter on.
    """
    return {"samples": MultilineOutputFormatter.format(p.samples)}


def processed_pep_filter(p, **kwargs) -> dict[str, str]:
    """
    Processed PEP filter, that returns the converted sample and subsample tables.

    This filter can return the tables as a table or a document.

    Args:
        p: A Project to run filter on.
        **kwargs: Optional keyword arguments including samples_as_objects
            and subsamples_as_objects flags to write as a table.
    """
    # get params
    samples_as_objects = kwargs.get("samples_as_objects")
    subsamples_as_objects = kwargs.get("subsamples_as_objects")

    prj_repr = p.config

    return {
        "project": str(prj_repr),
        "samples": (
            str(p.samples) if samples_as_objects else str(p.sample_table.to_csv())
        ),
        "subsamples": (
            str(p.subsamples)
            if subsamples_as_objects
            else str(p.subsample_table.to_csv())
        ),
    }



"""Exceptions for specific eido issues."""

from abc import ABCMeta

__all__ = [
    "EidoFilterError",
    "EidoSchemaInvalidError",
    "EidoValidationError",
    "PathAttrNotFoundError",
]


class EidoException(Exception):
    """Base type for custom package errors."""

    __metaclass__ = ABCMeta


class PathAttrNotFoundError(EidoException):
    """Path-like argument does not exist."""

    def __init__(self, key):
        super(PathAttrNotFoundError, self).__init__(key)


class EidoSchemaInvalidError(EidoException):
    """Schema does not comply to eido-specific requirements."""

    def __init__(self, key):
        super(EidoSchemaInvalidError, self).__init__(key)


class EidoFilterError(EidoException):
    """Issue with the PEP filter."""

    def __init__(self, key):
        super(EidoFilterError, self).__init__(key)


class EidoValidationError(EidoException):
    """Object was not validated successfully according to schema."""

    def __init__(self, message: str, errors_by_type: dict[str, list[dict[str, str]]]):
        super().__init__(message)
        self.errors_by_type = errors_by_type
        self.message = message

    def __str__(self):
        return f"EidoValidationError ({self.message}): {self.errors_by_type}"



import inspect
import os
import sys
from collections.abc import Callable
from importlib.metadata import entry_points
from logging import getLogger

from ..project import Project
from .exceptions import EidoFilterError

_LOGGER = getLogger(__name__)


def pep_conversion_plugins() -> dict[str, Callable]:
    """
    Plugins registered by entry points in the current Python env.

    Returns:
        Dict which keys are names of all possible hooks and values are dicts
        mapping registered functions names to their values

    Raises:
        EidoFilterError: If any of the filters has an invalid signature
    """
    plugins = {}
    for ep in entry_points(group="pep.filters"):
        plugin_fun = ep.load()
        if len(list(inspect.signature(plugin_fun).parameters)) != 2:
            raise EidoFilterError(
                f"Invalid filter plugin signature: {ep.name}. "
                f"Filter functions must take 2 arguments: peppy.Project and **kwargs"
            )
        plugins[ep.name] = plugin_fun
    return plugins


def convert_project(
    prj: Project, target_format: str, plugin_kwargs: dict | None = None
) -> dict[str, str]:
    """
    Convert a `peppy.Project` object to a selected format.

    Args:
        prj: A Project object to convert
        target_format: The format to convert the Project object to
        plugin_kwargs: Kwargs to pass to the plugin function

    Returns:
        Dictionary with conversion results

    Raises:
        EidoFilterError: If the requested filter is not defined
    """
    return run_filter(prj, target_format, plugin_kwargs=plugin_kwargs or dict())


def run_filter(
    prj: Project,
    filter_name: str,
    verbose: bool = True,
    plugin_kwargs: dict | None = None,
) -> dict[str, str]:
    """
    Run a selected filter on a peppy.Project object.

    Args:
        prj: A Project to run filter on
        filter_name: Name of the filter to run
        verbose: Whether to print output to stdout
        plugin_kwargs: Kwargs to pass to the plugin function

    Returns:
        Dictionary with conversion results

    Raises:
        EidoFilterError: If the requested filter is not defined
    """
    # convert to empty dictionary if no plugin_kwargs are passed
    plugin_kwargs = plugin_kwargs or dict()

    # get necessary objects
    installed_plugins = pep_conversion_plugins()
    installed_plugin_names = list(installed_plugins.keys())
    paths = plugin_kwargs.get("paths")
    env = plugin_kwargs.get("env")

    # set environment
    if env is not None:
        for var in env:
            os.environ[var] = env[var]

    # check for valid filter
    if filter_name not in installed_plugin_names:
        raise EidoFilterError(
            f"Requested filter ({filter_name}) not found. "
            f"Available: {', '.join(installed_plugin_names)}"
        )
    _LOGGER.info(f"Running plugin {filter_name}")
    func = installed_plugins[filter_name]

    # run filter
    conv_result = func(prj, **plugin_kwargs)

    # if paths supplied, write to disk
    if paths is not None:
        # map conversion result to the
        # specified path
        for result_key in conv_result:
            result_path = paths.get(result_key)
            if result_path is None:
                _LOGGER.warning(
                    f"Conversion plugin returned key that doesn't exist in specified paths: '{result_key}'."
                )
            else:
                # create path if it doesn't exist
                if not os.path.exists(result_path) and os.path.isdir(
                    os.path.dirname(result_path)
                ):
                    os.makedirs(os.path.dirname(result_path), exist_ok=True)
                save_result(result_path, conv_result[result_key])

    if verbose:
        for result_key in conv_result:
            sys.stdout.write(conv_result[result_key])

    return conv_result


def save_result(result_path: str, content: str) -> None:
    with open(result_path, "w") as f:
        f.write(content)


def get_available_pep_filters() -> list[str]:
    """
    Get a list of available target formats.

    Returns:
        A list of available formats
    """
    return list(pep_conversion_plugins().keys())



import json
import time

import requests
from pydantic import BaseModel

from ..helpers import MessageHandler, RequestManager
from ..pephub_oauth.const import PEPHUB_DEVICE_INIT_URI, PEPHUB_DEVICE_TOKEN_URI
from ..pephub_oauth.exceptions import (
    PEPHubResponseException,
    PEPHubTokenExchangeException,
)
from ..pephub_oauth.models import (
    InitializeDeviceCodeResponse,
    PEPHubDeviceTokenResponse,
)


class PEPHubAuth(RequestManager):
    """Class responsible for authorization to PEPhub."""

    def login_to_pephub(self):
        pephub_response = self._request_pephub_for_device_code()
        print(
            f"User verification code: {pephub_response.device_code}, please go to the website: "
            f"{pephub_response.auth_url} to authenticate."
        )

        # Sleep 2 minutes and then try 3 times exchange device code on token
        time.sleep(2)

        number_of_token_exchange_attempts = 3
        for i in range(number_of_token_exchange_attempts):
            try:
                user_token = self._exchange_device_code_on_token(
                    pephub_response.device_code
                )
            except PEPHubTokenExchangeException:
                time.sleep(2)
            else:
                print("Successfully logged in!")
                return user_token

        # If you didn't log in press enter to try again.
        input("If you logged in, press enter to continue...")
        try:
            user_token = self._exchange_device_code_on_token(
                pephub_response.device_code
            )
        except PEPHubTokenExchangeException:
            MessageHandler.print_warning("You are not logged in")
        else:
            MessageHandler.print_success("Successfully logged in!")
            return user_token

    def _request_pephub_for_device_code(self) -> InitializeDeviceCodeResponse:
        """Requests device code from pephub."""
        response = PEPHubAuth.send_request(
            method="POST",
            url=PEPHUB_DEVICE_INIT_URI,
            params=None,
            headers=None,
        )
        return self._handle_pephub_response(response, InitializeDeviceCodeResponse)

    def _exchange_device_code_on_token(self, device_code: str) -> str:
        """
        Send request with device code to pephub in order to exchange it on JWT.

        Args:
            device_code: device code that was generated by pephub
        """
        response = PEPHubAuth.send_request(
            method="POST",
            url=PEPHUB_DEVICE_TOKEN_URI,
            params=None,
            headers={"device-code": device_code},
        )
        pephub_token_response = self._handle_pephub_response(
            response, PEPHubDeviceTokenResponse
        )
        return pephub_token_response.jwt_token

    @staticmethod
    def _handle_pephub_response(
        response: requests.Response, model: type[BaseModel]
    ) -> BaseModel | InitializeDeviceCodeResponse | PEPHubDeviceTokenResponse:
        """
        Decode the response from PEPhub and pack the returned data into appropriate model.

        Args:
            response: Response from pephub
            model: Model that the data will be packed to.

        Returns:
            Response data as an instance of correct model.
        """
        if response.status_code == 401:
            raise PEPHubTokenExchangeException()
        if response.status_code != 200:
            raise PEPHubResponseException()
        try:
            content = json.loads(PEPHubAuth.decode_response(response))
        except json.JSONDecodeError:
            raise Exception("Something went wrong with PEPhub response")

        return model(**content)



import os
from enum import Enum

from pydantic import BaseModel, field_validator

PEPHUB_BASE_URL = os.getenv(
    "PEPHUB_BASE_URL", default="https://pephub-api.databio.org/"
)
# PEPHUB_BASE_URL = "http://0.0.0.0:8000/"
PEPHUB_PEP_API_BASE_URL = f"{PEPHUB_BASE_URL}api/v1/projects/"
PEPHUB_PEP_SEARCH_URL = f"{PEPHUB_BASE_URL}api/v1/namespaces/{{namespace}}/projects"
PEPHUB_PUSH_URL = f"{PEPHUB_BASE_URL}api/v1/namespaces/{{namespace}}/projects/json"

PEPHUB_SAMPLE_URL = f"{PEPHUB_BASE_URL}api/v1/projects/{{namespace}}/{{project}}/samples/{{sample_name}}"
PEPHUB_VIEW_URL = (
    f"{PEPHUB_BASE_URL}api/v1/projects/{{namespace}}/{{project}}/views/{{view_name}}"
)
PEPHUB_VIEW_SAMPLE_URL = f"{PEPHUB_BASE_URL}api/v1/projects/{{namespace}}/{{project}}/views/{{view_name}}/{{sample_name}}"


class RegistryPath(BaseModel):
    protocol: str | None = None
    namespace: str
    item: str
    subitem: str | None = None
    tag: str | None = "default"

    @field_validator("tag")
    def tag_should_not_be_none(cls, v):
        return v or "default"


class ResponseStatusCodes(int, Enum):
    OK = 200
    ACCEPTED = 202
    UNAUTHORIZED = 401
    FORBIDDEN = 403
    NOT_EXIST = 404
    CONFLICT = 409
    INTERNAL_ERROR = 500


USER_DATA_FILE_NAME = "jwt.txt"
HOME_PATH = os.getenv("HOME")
if not HOME_PATH:
    HOME_PATH = os.path.expanduser("~")
PATH_TO_FILE_WITH_JWT = os.path.join(HOME_PATH, ".pephubclient/") + USER_DATA_FILE_NAME



"""Constant variables for eido package."""

LOGGING_LEVEL = "INFO"
PKG_NAME = "eido"
INSPECT_CMD = "inspect"
VALIDATE_CMD = "validate"
CONVERT_CMD = "convert"
FILTERS_CMD = "filters"
SUBPARSER_MSGS = {
    VALIDATE_CMD: "Validate a PEP or its components",
    INSPECT_CMD: "Inspect a PEP",
    CONVERT_CMD: "Convert PEP format using filters",
}
PROP_KEY = "properties"

SAMPLES_KEY = "samples"

TANGIBLE_KEY = "tangible"
SIZING_KEY = "sizing"

# sample schema input validation key names, these values are required by looper
# to refer to the dict values
MISSING_KEY = "missing"
REQUIRED_INPUTS_KEY = "required_inputs"
ALL_INPUTS_KEY = "all_inputs"
INPUT_FILE_SIZE_KEY = "input_file_size"

# groups of constants
GENERAL = [
    "LOGGING_LEVEL",
    "PKG_NAME",
    "INSPECT_CMD",
    "VALIDATE_CMD",
    "CONVERT_CMD",
    "FILTERS_CMD",
    "SUBPARSER_MSGS",
]

SCHEMA_SECTIONS = ["PROP_KEY", "TANGIBLE_KEY", "SIZING_KEY"]

SCHEMA_VALIDATION_KEYS = [
    "MISSING_KEY",
    "REQUIRED_INPUTS_KEY",
    "ALL_INPUTS_KEY",
    "INPUT_FILE_SIZE_KEY",
]

__all__ = GENERAL + SCHEMA_SECTIONS + SCHEMA_VALIDATION_KEYS



from unittest.mock import Mock

import pytest

from peppy.pephubclient.exceptions import ResponseError
from peppy.pephubclient.helpers import is_registry_path
from peppy.pephubclient.pephub_oauth.models import InitializeDeviceCodeResponse
from peppy.pephubclient.pephubclient import PEPHubClient


@pytest.fixture()
def device_code_return():
    device_code = "asdf2345"
    return InitializeDeviceCodeResponse(
        device_code=device_code,
        auth_url=f"any_base_url/auth/device/login/{device_code}",
    )


class TestSmoke:
    def test_login(self, mocker, device_code_return, test_jwt):
        """
        Test if device login request was sent to pephub
        """
        requests_mock = mocker.patch(
            "requests.request",
            return_value=Mock(content=device_code_return, status_code=200),
        )
        pephub_response_mock = mocker.patch(
            "peppy.pephubclient.pephub_oauth.pephub_oauth.PEPHubAuth._handle_pephub_response",
            return_value=device_code_return,
        )

        pephub_exchange_code_mock = mocker.patch(
            "peppy.pephubclient.pephub_oauth.pephub_oauth.PEPHubAuth._exchange_device_code_on_token",
            return_value=test_jwt,
        )

        pathlib_mock = mocker.patch(
            "peppy.pephubclient.files_manager.FilesManager.save_jwt_data_to_file"
        )

        PEPHubClient().login()

        assert requests_mock.called
        assert pephub_response_mock.called
        assert pephub_exchange_code_mock.called
        assert pathlib_mock.called

    def test_logout(self, mocker):
        os_remove_mock = mocker.patch("os.remove")
        PEPHubClient().logout()

        assert os_remove_mock.called

    def test_pull(self, mocker, test_jwt, test_raw_pep_return):
        jwt_mock = mocker.patch(
            "peppy.pephubclient.files_manager.FilesManager.load_jwt_data_from_file",
            return_value=test_jwt,
        )
        requests_mock = mocker.patch(
            "requests.request",
            return_value=Mock(content="some return", status_code=200),
        )
        mocker.patch(
            "peppy.pephubclient.helpers.RequestManager.decode_response",
            return_value=test_raw_pep_return,
        )
        save_yaml_mock = mocker.patch(
            "peppy.pephubclient.files_manager.FilesManager.save_yaml"
        )
        save_sample_mock = mocker.patch(
            "peppy.pephubclient.files_manager.FilesManager.save_pandas"
        )
        mocker.patch(
            "peppy.pephubclient.files_manager.FilesManager.create_project_folder"
        )

        PEPHubClient().pull("some/project")

        assert jwt_mock.called
        assert requests_mock.called
        assert save_yaml_mock.called
        assert save_sample_mock.called

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "File does not exist, or you are unauthorized.",
            ),
            (
                500,
                "Internal server error. Unexpected return value. Error: 500",
            ),
        ],
    )
    def test_pull_with_pephub_error_response(
        self, mocker, test_jwt, status_code, expected_error_message
    ):
        mocker.patch(
            "peppy.pephubclient.files_manager.FilesManager.load_jwt_data_from_file",
            return_value=test_jwt,
        )
        mocker.patch(
            "requests.request",
            return_value=Mock(
                content=b'{"detail": "Some error message"}', status_code=status_code
            ),
        )

        with pytest.raises(ResponseError) as e:
            PEPHubClient().pull("some/project")

        assert e.value.message == expected_error_message

    def test_push(self, mocker, test_jwt, SAMPLE_PEP):
        requests_mock = mocker.patch(
            "requests.request", return_value=Mock(status_code=202)
        )

        PEPHubClient().push(
            SAMPLE_PEP,
            namespace="s_name",
            name="name",
        )

        assert requests_mock.called

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                409,
                "Project already exists. Set force to overwrite project.",
            ),
            (
                401,
                "Unauthorized! Failure in uploading project.",
            ),
            (233, "Unexpected Response Error."),
        ],
    )
    def test_push_with_pephub_error_response(
        self, mocker, status_code, expected_error_message, SAMPLE_PEP
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().push(
                SAMPLE_PEP,
                namespace="s_name",
                name="name",
            )

    def test_search_prj(self, mocker):
        return_value = {
            "count": 1,
            "limit": 100,
            "offset": 0,
            "results": [
                {
                    "namespace": "namespace1",
                    "name": "basic",
                    "tag": "default",
                    "is_private": False,
                    "number_of_samples": 2,
                    "description": "None",
                    "last_update_date": "2023-08-27 19:07:31.552861+00:00",
                    "submission_date": "2023-08-27 19:07:31.552858+00:00",
                    "digest": "08cbcdbf4974fc84bee824c562b324b5",
                    "pep_schema": "random_schema_name",
                    "pop": False,
                    "stars_number": 0,
                    "forked_from": None,
                }
            ],
            "session_info": None,
            "can_edit": False,
        }
        mocker.patch(
            "requests.request",
            return_value=Mock(content=return_value, status_code=200),
        )
        mocker.patch(
            "peppy.pephubclient.helpers.RequestManager.decode_response",
            return_value=return_value,
        )

        return_value = PEPHubClient().find_project(namespace="namespace1")
        assert return_value.count == 1
        assert len(return_value.results) == 1


class TestHelpers:
    @pytest.mark.parametrize(
        "input_str, expected_output",
        [
            (
                "databio/pep:default",
                True,
            ),
            (
                "pephub.databio.org::databio/pep:default",
                True,
            ),
            (
                "pephub.databio.org://databio/pep:default",
                True,
            ),
            (
                "databio/pep",
                True,
            ),
            (
                "databio/pep/default",
                False,
            ),
            (
                "some/random/path/to.yaml",
                False,
            ),
            (
                "path_to.csv",
                False,
            ),
            (
                "this/is/path/to.csv",
                False,
            ),
        ],
    )
    def test_is_registry_path(self, input_str, expected_output):
        assert is_registry_path(input_str) is expected_output


class TestSamples:
    def test_get(self, mocker):
        return_value = {
            "genome": "phc_test1",
            "sample_type": "phc_test",
            "sample_name": "gg1",
        }
        mocker.patch(
            "requests.request",
            return_value=Mock(content=return_value, status_code=200),
        )
        mocker.patch(
            "peppy.pephubclient.helpers.RequestManager.decode_response",
            return_value=return_value,
        )
        return_value = PEPHubClient().sample.get(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
        )
        assert return_value == return_value

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "Sample does not exist.",
            ),
            (
                500,
                "Internal server error. Unexpected return value.",
            ),
            (
                403,
                "Unexpected return value. Error: 403",
            ),
        ],
    )
    def test_sample_get_with_pephub_error_response(
        self, mocker, status_code, expected_error_message
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().sample.get(
                "test_namespace",
                "test_name",
                "default",
                "gg1",
            )

    @pytest.mark.parametrize(
        "prj_dict",
        [
            {"genome": "phc_test1", "sample_type": "phc_test", "sample_name": "gg1"},
            {"genome": "phc_test1", "sample_type": "phc_test"},
        ],
    )
    def test_create(self, mocker, prj_dict):
        return_value = prj_dict
        mocker_obj = mocker.patch(
            "requests.request",
            return_value=Mock(content=return_value, status_code=202),
        )

        PEPHubClient().sample.create(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
            sample_dict=return_value,
        )
        assert mocker_obj.called

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "does not exist.",
            ),
            (
                409,
                "already exists. Set overwrite to True to overwrite sample.",
            ),
            (
                500,
                "Unexpected return value.",
            ),
        ],
    )
    def test_sample_create_with_pephub_error_response(
        self, mocker, status_code, expected_error_message
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().sample.create(
                "test_namespace",
                "test_name",
                "default",
                "gg1",
                sample_dict={
                    "genome": "phc_test1",
                    "sample_type": "phc_test",
                    "sample_name": "gg1",
                },
            )

    def test_delete(self, mocker):
        mocker_obj = mocker.patch(
            "requests.request",
            return_value=Mock(status_code=202),
        )

        PEPHubClient().sample.remove(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
        )
        assert mocker_obj.called

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "does not exist.",
            ),
            (
                500,
                "Unexpected return value.",
            ),
        ],
    )
    def test_sample_delete_with_pephub_error_response(
        self, mocker, status_code, expected_error_message
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().sample.remove(
                "test_namespace",
                "test_name",
                "default",
                "gg1",
            )

    def test_update(self, mocker):
        mocker_obj = mocker.patch(
            "requests.request",
            return_value=Mock(status_code=202),
        )

        PEPHubClient().sample.update(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
            sample_dict={
                "genome": "phc_test1",
                "sample_type": "phc_test",
                "new_col": "column",
            },
        )
        assert mocker_obj.called

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "does not exist.",
            ),
            (
                500,
                "Unexpected return value.",
            ),
        ],
    )
    def test_sample_update_with_pephub_error_response(
        self, mocker, status_code, expected_error_message
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().sample.update(
                "test_namespace",
                "test_name",
                "default",
                "gg1",
                sample_dict={
                    "genome": "phc_test1",
                    "sample_type": "phc_test",
                    "new_col": "column",
                },
            )


class TestViews:
    def test_get(self, mocker, test_raw_pep_return):
        return_value = test_raw_pep_return
        mocker.patch(
            "requests.request",
            return_value=Mock(content=return_value, status_code=200),
        )
        mocker.patch(
            "peppy.pephubclient.helpers.RequestManager.decode_response",
            return_value=return_value,
        )

        return_value = PEPHubClient().view.get(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
        )
        assert return_value == return_value

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "does not exist.",
            ),
            (
                500,
                "Internal server error.",
            ),
        ],
    )
    def test_view_get_with_pephub_error_response(
        self, mocker, status_code, expected_error_message
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().view.get(
                "test_namespace",
                "test_name",
                "default",
                "gg1",
            )

    def test_create(self, mocker):
        mocker_obj = mocker.patch(
            "requests.request",
            return_value=Mock(status_code=202),
        )

        PEPHubClient().view.create(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
            sample_list=["sample1", "sample2"],
        )
        assert mocker_obj.called

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "does not exist.",
            ),
            (
                409,
                "already exists in the project.",
            ),
        ],
    )
    def test_view_create_with_pephub_error_response(
        self, mocker, status_code, expected_error_message
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().view.create(
                "test_namespace",
                "test_name",
                "default",
                "gg1",
                sample_list=["sample1", "sample2"],
            )

    def test_delete(self, mocker):
        mocker_obj = mocker.patch(
            "requests.request",
            return_value=Mock(status_code=202),
        )

        PEPHubClient().view.delete(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
        )
        assert mocker_obj.called

    @pytest.mark.parametrize(
        "status_code, expected_error_message",
        [
            (
                404,
                "does not exist.",
            ),
            (
                401,
                "You are unauthorized to delete this view.",
            ),
        ],
    )
    def test_view_delete_with_pephub_error_response(
        self, mocker, status_code, expected_error_message
    ):
        mocker.patch("requests.request", return_value=Mock(status_code=status_code))
        with pytest.raises(ResponseError, match=expected_error_message):
            PEPHubClient().view.delete(
                "test_namespace",
                "test_name",
                "default",
                "gg1",
            )

    def test_add_sample(self, mocker):
        mocker_obj = mocker.patch(
            "requests.request",
            return_value=Mock(status_code=202),
        )

        PEPHubClient().view.add_sample(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
            "sample1",
        )
        assert mocker_obj.called

    def test_delete_sample(self, mocker):
        mocker_obj = mocker.patch(
            "requests.request",
            return_value=Mock(status_code=202),
        )

        PEPHubClient().view.remove_sample(
            "test_namespace",
            "test_name",
            "default",
            "gg1",
            "sample1",
        )
        assert mocker_obj.called


###


# test add sample:
# 1. add correct 202
# 2. add existing 409
# 3. add with sample_name
# 4. add without sample_name
# 5. add with overwrite
# 6. add to nonexistent project 404

# delete sample:
# 1. delete existing 202
# 2. delete nonexistent 404

# get sample:
# 1. get existing 200
# 2. get nonexistent 404
# 3. get with raw 200
# 4. get from nonexistent project 404

# update sample:
# 1. update existing 202
# 2. update nonexistent sample 404
# 3. update nonexistent project 404



from abc import ABC, abstractmethod
from collections.abc import Iterable

from ..sample import Sample


class BaseOutputFormatter(ABC):
    @staticmethod
    @abstractmethod
    def format(samples: list[Sample]) -> str:
        """Convert the samples to correct format."""
        pass


class MultilineOutputFormatter(BaseOutputFormatter):
    @staticmethod
    def format(samples: list[Sample]) -> str:
        output_rows = []
        sample_attributes = [
            attribute
            for attribute in samples[0].keys()
            if not attribute.startswith("_") and not attribute == "subsample_name"
        ]
        header = MultilineOutputFormatter._get_header(sample_attributes)

        for sample in samples:
            attribute_with_multiple_properties = MultilineOutputFormatter._get_the_name_of_the_first_attribute_with_multiple_properties(
                sample, sample_attributes
            )
            if attribute_with_multiple_properties:
                sample_rows = MultilineOutputFormatter._split_sample_to_multiple_rows(
                    sample, sample_attributes, attribute_with_multiple_properties
                )
                output_rows.extend(sample_rows)
            else:
                one_sample_row = MultilineOutputFormatter._convert_sample_to_row(
                    sample, sample_attributes
                )
                output_rows.append(one_sample_row)

        return "\n".join(header + output_rows) + "\n"

    @staticmethod
    def _get_header(header_column_names: list[str]) -> list[str]:
        return [",".join(header_column_names)]

    @staticmethod
    def _get_the_name_of_the_first_attribute_with_multiple_properties(
        sample: Sample, sample_attributes: list[str]
    ) -> str | None:
        for attribute in sample_attributes:
            if MultilineOutputFormatter._sample_attribute_is_list(sample, attribute):
                return attribute

    @staticmethod
    def _split_sample_to_multiple_rows(
        sample: Sample, sample_attributes: list, attribute_with_multiple_properties: str
    ) -> Iterable[str]:
        """
        If one sample object contains array properties instead of single value, then it will be converted
        to multiple rows.

        Args:
            sample: Sample from project.
            sample_attributes: List of all sample properties names (name of columns from sample_table).

        Returns:
            List of rows created from given sample object.
        """
        number_of_samples_after_split = len(
            getattr(sample, attribute_with_multiple_properties)
        )
        sample_rows_after_split = []

        for sample_index in range(number_of_samples_after_split):
            sample_row = MultilineOutputFormatter._convert_sample_to_row(
                sample, sample_attributes, sample_index
            )
            sample_rows_after_split.append(sample_row)

        return sample_rows_after_split

    @staticmethod
    def _convert_sample_to_row(
        sample: Sample, sample_attributes: list, sample_index: int = 0
    ) -> str:
        """
        Converts single sample object to CSV row.

        Some samples have a list of values instead of single value for given attribute (column), and
        sample_index indicates index of the value that will be used to create a row. For samples that don't
        have any attributes with given names this will always be zero.

        Args:
            sample: Single sample object.
            sample_attributes: Array of all attributes names (column names) for given sample.
            sample_index: Number indicating which value will be used to create row. Some samples

        Returns:
            Representation of sample as a CSV row.
        """
        sample_row = []

        for attribute in sample_attributes:
            if (
                MultilineOutputFormatter._sample_attribute_is_list(sample, attribute)
                and sample[attribute]
            ):
                value = sample[attribute][sample_index]
            else:
                value = sample.get(attribute)

            sample_row.append(value or "")

        return ",".join(sample_row)

    @staticmethod
    def _sample_attribute_is_list(sample: Sample, attribute: str) -> bool:
        return isinstance(getattr(sample, attribute, ""), list)


class SampleSubsampleOutputFormatter(BaseOutputFormatter):
    def format(self, samples: list[Sample]) -> str:
        pass



from collections.abc import Iterator, MutableMapping
from typing import Any


class SimpleAttMap(MutableMapping):
    """
    Simplified the AttrMap class, which enables storing key-value pairs in
    a dictionary-like structure.
    It allows assigning and accessing the values both through attributes and items.
    In most cases used as SuperClass.
    """

    def __init__(self) -> None:
        super(SimpleAttMap, self).__init__()
        super(SimpleAttMap, self).__setattr__("_mapped_attr", {})

    def __delitem__(self, key: str) -> None:
        value = self[key]
        del self._mapped_attr[key]
        self.pop(value, None)

    def __setitem__(self, item: str, value: Any) -> None:
        self._try_touch_samples()
        self._mapped_attr[item] = value

    def __getitem__(self, item: str) -> Any:
        return self._mapped_attr[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self._mapped_attr)

    def __len__(self) -> int:
        return len(self._mapped_attr)

    def __contains__(self, key: object) -> bool:
        return key in list(self.keys())

    def __delattr__(self, key: str) -> None:
        del self[key]

    def __setattr__(self, item: str, value: Any) -> None:
        self._mapped_attr[item] = value

    def __getattr__(self, item: str) -> Any:
        try:
            return self._mapped_attr[item]
        except KeyError:
            raise AttributeError(f"Attribute not found: {item}")

    def __eq__(self, other: "SimpleAttMap") -> bool:
        return self._mapped_attr == other._mapped_attr

    @property
    def attributes(self) -> dict:
        return self._mapped_attr



import datetime

from pydantic import BaseModel, ConfigDict, Field, field_validator

from ..const import CONFIG_KEY, SAMPLE_RAW_DICT_KEY, SUBSAMPLE_RAW_LIST_KEY


class ProjectDict(BaseModel):
    """
    Project dict (raw) model
    """

    config: dict = Field(alias=CONFIG_KEY)
    subsample_list: list | None = Field(alias=SUBSAMPLE_RAW_LIST_KEY)
    sample_list: list = Field(alias=SAMPLE_RAW_DICT_KEY)

    model_config = ConfigDict(populate_by_name=True, extra="allow")


class ProjectUploadData(BaseModel):
    """
    Model used in post request to upload project
    """

    pep_dict: ProjectDict
    tag: str | None = "default"
    is_private: bool | None = False
    overwrite: bool | None = False

    @field_validator("tag")
    def tag_should_not_be_none(cls, v):
        return v or "default"


class ProjectAnnotationModel(BaseModel):
    namespace: str
    name: str
    tag: str
    is_private: bool
    number_of_samples: int
    description: str
    last_update_date: datetime.datetime
    submission_date: datetime.datetime
    digest: str
    pep_schema: str | int | None = None
    pop: bool = False
    stars_number: int | None = 0
    forked_from: str | None = None


class SearchReturnModel(BaseModel):
    count: int
    limit: int
    offset: int
    results: list[ProjectAnnotationModel]



import os
import tempfile

import pytest

from peppy import Project

__author__ = "Michal Stolarczyk"
__email__ = "michal.stolarczyk@nih.gov"

EXAMPLE_TYPES = [
    "basic",
    "derive",
    "imply",
    "append",
    "amendments1",
    "amendments2",
    "derive_imply",
    "duplicate",
    "imports",
    "subtable1",
    "subtable2",
    "subtable3",
    "subtable4",
    "subtable5",
    "remove",
]


class TestSample:
    @pytest.mark.parametrize("example_pep_cfg_path", ["basic"], indirect=True)
    def test_serialization(self, example_pep_cfg_path):
        """
        Verify that Project object is successfully created for every example PEP
        """
        td = tempfile.mkdtemp()
        fn = os.path.join(td, "serialized_sample.yaml")
        p = Project(cfg=example_pep_cfg_path)
        sample = p.samples[0]
        sample["set"] = set("set")
        sample["dict"] = dict({"dict": "dict"})
        sample["list"] = list(["list"])
        sample.to_yaml(fn)
        with open(fn, "r") as f:
            contents = f.read()
        assert "set" in contents
        assert "dict" in contents
        assert "list" in contents

    @pytest.mark.parametrize("example_pep_cfg_path", EXAMPLE_TYPES, indirect=True)
    def test_str_repr_correctness(self, example_pep_cfg_path):
        """
        Verify that the missing amendment request raises correct exception
        """
        p = Project(cfg=example_pep_cfg_path)
        for sample in p.samples:
            str_repr = sample.__str__(max_attr=100)
            assert example_pep_cfg_path in str_repr
            assert "Sample '{}'".format(sample["sample_name"]) in str_repr

    @pytest.mark.parametrize("example_pep_cfg_path", EXAMPLE_TYPES, indirect=True)
    def test_sample_to_yaml_no_path(self, example_pep_cfg_path):
        """
        Verify that to_yaml returns representation without requiring a path.
        """
        p = Project(cfg=example_pep_cfg_path)
        for sample in p.samples:
            yaml_repr = sample.to_yaml()
            assert "sample_name" in yaml_repr

    @pytest.mark.parametrize("example_pep_cfg_path", ["basic"], indirect=True)
    def test_sheet_dict_excludes_private_attrs(self, example_pep_cfg_path):
        """
        Verify that sheet dict includes only original Sample attributes
        """
        p = Project(cfg=example_pep_cfg_path)
        for sample in p.samples:
            assert len(sample.get_sheet_dict()) == len(p.sample_table.columns)

    @pytest.mark.parametrize("example_pep_cfg_path", ["basic"], indirect=True)
    def test_pickle_in_samples(self, example_pep_cfg_path):
        import pickle

        p = Project(cfg=example_pep_cfg_path)
        for sample in p.samples:
            pickled_data = pickle.dumps(sample)
            unpickled_sample = pickle.loads(pickled_data)

            assert sample.to_dict() == unpickled_sample.to_dict()

    @pytest.mark.parametrize("example_pep_cfg_path", ["basic"], indirect=True)
    def test_equals_samples(self, example_pep_cfg_path):
        p1 = Project(cfg=example_pep_cfg_path)
        p2 = Project(cfg=example_pep_cfg_path)
        s1 = p1.samples[0]
        s2 = p2.samples[0]

        assert s1 == s2

    @pytest.mark.parametrize("example_pep_cfg_path", ["basic"], indirect=True)
    def test_not_equals_samples(self, example_pep_cfg_path):
        p1 = Project(cfg=example_pep_cfg_path)
        p2 = Project(cfg=example_pep_cfg_path)
        s1 = p1.samples[0]
        s2 = p2.samples[0]
        s3 = p2.samples[1]

        s2.new = "something"
        assert not s1 == s2
        assert not s1 == s3



"""auth exceptions"""


class PEPHubResponseException(Exception):
    """Request response exception. Used when response != 200."""

    def __init__(self, reason: str = ""):
        super(PEPHubResponseException, self).__init__(reason)


class PEPHubTokenExchangeException(Exception):
    """Exception in exchanging device code on token == 400."""

    def __init__(self, reason: str = ""):
        super(PEPHubTokenExchangeException, self).__init__(reason)



import json
import os
from collections.abc import Callable
from typing import Any
from urllib.parse import urlencode

import pandas as pd
import requests
import yaml
from pydantic import ValidationError
from requests.exceptions import ConnectionError
from ubiquerg import parse_registry_path

from ..const import (
    CFG_SAMPLE_TABLE_KEY,
    CFG_SUBSAMPLE_TABLE_KEY,
    CONFIG_KEY,
    DESC_KEY,
    NAME_KEY,
    SAMPLE_RAW_DICT_KEY,
    SUBSAMPLE_RAW_LIST_KEY,
)
from ..project import Project
from .constants import RegistryPath
from .exceptions import PEPExistsError, ResponseError
from .files_manager import FilesManager
from .models import ProjectDict


class RequestManager:
    @staticmethod
    def send_request(
        method: str,
        url: str,
        headers: dict | None = None,
        cookies: dict | None = None,
        params: dict | None = None,
        json: dict | list | None = None,
    ) -> requests.Response:
        request_return = requests.request(
            method=method,
            url=url,
            verify=False,
            cookies=cookies,
            headers=headers,
            params=params,
            json=json,
            timeout=10,
        )
        if request_return.status_code == 401:
            if (
                RequestManager.decode_response(request_return, output_json=True).get(
                    "detail"
                )
                == "JWT has expired"
            ):
                raise ResponseError("JWT has expired. Please log in again.")
        return request_return

    @staticmethod
    def decode_response(
        response: requests.Response, encoding: str = "utf-8", output_json: bool = False
    ) -> str | dict:
        """
        Decode the response from PEPhub and pack the returned data into appropriate model.

        Args:
            response: Response from PEPhub.
            encoding: Response encoding [Default: utf-8]
            output_json: If True, return response in json format

        Returns:
            Response data as an instance of correct model.
        """

        try:
            if output_json:
                return response.json()
            else:
                return response.content.decode(encoding)
        except json.JSONDecodeError as err:
            raise ResponseError(f"Error in response encoding format: {err}")

    @staticmethod
    def parse_query_param(pep_variables: dict) -> str:
        """
        Grab all the variables passed by user (if any) and parse them to match the format
        specified by PEPhub API for query parameters.

        Args:
            pep_variables: dict of query parameters

        Returns:
            PEPHubClient variables transformed into string in correct format.
        """
        return "?" + urlencode(pep_variables)

    @staticmethod
    def parse_header(jwt_data: str | None = None) -> dict:
        """
        Create Authorization header.

        Args:
            jwt_data: jwt string

        Returns:
            Authorization dict.
        """
        if jwt_data:
            return {"Authorization": jwt_data}
        else:
            return {}


class MessageHandler:
    """Class holding print function in different colors."""

    RED = 9
    YELLOW = 11
    GREEN = 40

    @staticmethod
    def print_error(text: str) -> None:
        print(f"\033[38;5;9m{text}\033[0m")

    @staticmethod
    def print_success(text: str) -> None:
        print(f"\033[38;5;40m{text}\033[0m")

    @staticmethod
    def print_warning(text: str) -> None:
        print(f"\033[38;5;11m{text}\033[0m")


def call_client_func(func: Callable[..., Any], **kwargs) -> Any:
    """
    Catch exceptions in functions called through cli.

    Args:
        func: The function to call.
        **kwargs: The keyword arguments to pass to the function.

    Returns:
        The result of the function call.
    """

    try:
        func(**kwargs)
    except ConnectionError as err:
        MessageHandler.print_error(f"Failed to connect to server. Try later. {err}")
    except ResponseError as err:
        MessageHandler.print_error(f"{err}")
    except PEPExistsError as err:
        MessageHandler.print_warning(f"PEP already exists. {err}")
    except OSError as err:
        MessageHandler.print_error(f"{err}")


def is_registry_path(input_string: str) -> bool:
    """
    Check if input is a registry path to pephub.

    Args:
        input_string: path to the PEP (or registry path)

    Returns:
        True if input is a registry path.
    """
    if input_string.endswith(".yaml"):
        return False
    try:
        RegistryPath(**parse_registry_path(input_string))
    except (ValidationError, TypeError):
        return False
    return True


def unwrap_registry_path(input_string: str) -> RegistryPath:
    """
    Unwrap registry path from string.

    Args:
        input_string: path to the PEP (or registry path)

    Returns:
        RegistryPath object.
    """
    return RegistryPath(**parse_registry_path(input_string))


def _build_filename(registry_path: RegistryPath) -> str:
    """
    Takes query string and creates output filename to save the project to.

    Args:
        registry_path: Query string that was used to find the project.

    Returns:
        Filename uniquely identifying the project.
    """
    filename = "_".join(filter(bool, [registry_path.namespace, registry_path.item]))
    if registry_path.tag:
        filename += f"_{registry_path.tag}"
    return filename


def _save_zip_pep(project: dict, zip_filepath: str, force: bool = False) -> None:
    """
    Zip and save a project.

    Args:
        project: peppy project to zip
        zip_filepath: path to save zip file
        force: overwrite project if exists
    """

    content_to_zip = {}
    config = project[CONFIG_KEY]
    project_name = config[NAME_KEY]

    if project[SAMPLE_RAW_DICT_KEY] is not None:
        config[CFG_SAMPLE_TABLE_KEY] = ["sample_table.csv"]
        content_to_zip["sample_table.csv"] = pd.DataFrame(
            project[SAMPLE_RAW_DICT_KEY]
        ).to_csv(index=False)

    if project[SUBSAMPLE_RAW_LIST_KEY] is not None:
        if not isinstance(project[SUBSAMPLE_RAW_LIST_KEY], list):
            config[CFG_SUBSAMPLE_TABLE_KEY] = ["subsample_table1.csv"]
            content_to_zip["subsample_table1.csv"] = pd.DataFrame(
                project[SUBSAMPLE_RAW_LIST_KEY]
            ).to_csv(index=False)
        else:
            config[CFG_SUBSAMPLE_TABLE_KEY] = []
            for number, file in enumerate(project[SUBSAMPLE_RAW_LIST_KEY]):
                file_name = f"subsample_table{number + 1}.csv"
                config[CFG_SUBSAMPLE_TABLE_KEY].append(file_name)
                content_to_zip[file_name] = pd.DataFrame(file).to_csv(index=False)

    content_to_zip[f"{project_name}_config.yaml"] = yaml.dump(config, indent=4)
    FilesManager.save_zip_file(content_to_zip, file_path=zip_filepath, force=force)

    MessageHandler.print_success(f"Project was saved successfully -> {zip_filepath}")
    return None


def _save_unzipped_pep(
    project_dict: dict, folder_path: str, force: bool = False
) -> None:
    """
    Save unzipped project to specified folder.

    Args:
        project_dict: raw pep project
        folder_path: path to save project
        force: overwrite project if exists
    """

    def full_path(fn: str) -> str:
        return os.path.join(folder_path, fn)

    project_name = project_dict[CONFIG_KEY][NAME_KEY]
    sample_table_filename = "sample_table.csv"
    yaml_full_path = full_path(f"{project_name}_config.yaml")
    sample_full_path = full_path(sample_table_filename)
    if not force:
        extant = [p for p in [yaml_full_path, sample_full_path] if os.path.isfile(p)]
        if extant:
            raise PEPExistsError(f"{len(extant)} file(s) exist(s): {', '.join(extant)}")

    config_dict = project_dict.get(CONFIG_KEY)
    config_dict[NAME_KEY] = project_name
    config_dict[DESC_KEY] = project_dict[CONFIG_KEY][DESC_KEY]
    config_dict["sample_table"] = sample_table_filename

    sample_pandas = pd.DataFrame(project_dict.get(SAMPLE_RAW_DICT_KEY, {}))

    subsample_list = [
        pd.DataFrame(sub_a) for sub_a in project_dict.get(SUBSAMPLE_RAW_LIST_KEY) or []
    ]

    filenames = []
    for idx, subsample in enumerate(subsample_list):
        fn = f"subsample_table{idx + 1}.csv"
        filenames.append(fn)
        FilesManager.save_pandas(subsample, full_path(fn), not_force=False)
    config_dict["subsample_table"] = filenames

    FilesManager.save_yaml(config_dict, yaml_full_path, not_force=False)
    FilesManager.save_pandas(sample_pandas, sample_full_path, not_force=False)

    if config_dict.get("subsample_table"):
        for number, subsample in enumerate(subsample_list):
            FilesManager.save_pandas(
                subsample,
                os.path.join(folder_path, config_dict["subsample_table"][number]),
                not_force=False,
            )

    MessageHandler.print_success(f"Project was saved successfully -> {folder_path}")
    return None


def save_pep(
    project: dict | Project,
    reg_path: str = None,
    force: bool = False,
    project_path: str | None = None,
    zip: bool = False,
) -> None:
    """
    Save project locally.

    Args:
        project: PEP dictionary (raw project)
        reg_path: Project registry path in PEPhub (e.g. databio/base:default). If not
            provided, folder will be created with just project name.
        force: overwrite project if exists
        project_path: Path where project will be saved. By default, it will be saved in
            current directory.
        zip: If True, save project as zip file
    """
    if isinstance(project, Project):
        project = project.to_dict(extended=True, orient="records")

    project = ProjectDict(**project).model_dump(by_alias=True)

    if not project_path:
        project_path = os.getcwd()

    if reg_path:
        file_name = _build_filename(RegistryPath(**parse_registry_path(reg_path)))
    else:
        file_name = project[CONFIG_KEY][NAME_KEY]

    if zip:
        _save_zip_pep(
            project,
            zip_filepath=f"{os.path.join(project_path, file_name)}.zip",
            force=force,
        )
        return None

    folder_path = FilesManager.create_project_folder(
        parent_path=project_path, folder_name=file_name
    )
    _save_unzipped_pep(project, folder_path, force=force)



import os

import pytest

from peppy import Project


@pytest.fixture
def data_path():
    return os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "data", "eidodata"
    )


@pytest.fixture
def schemas_path(data_path):
    return os.path.join(data_path, "schemas")


@pytest.fixture
def peps_path(data_path):
    return os.path.join(data_path, "peps")


@pytest.fixture
def project_file_path(peps_path):
    return os.path.join(peps_path, "test_pep", "test_cfg.yaml")


@pytest.fixture
def project_table_path(peps_path):
    return os.path.join(peps_path, "test_pep", "test_sample_table.csv")


@pytest.fixture
def project_object(project_file_path):
    return Project(project_file_path)


@pytest.fixture
def schema_file_path(schemas_path):
    return os.path.join(schemas_path, "test_schema.yaml")


@pytest.fixture
def schema_samples_file_path(schemas_path):
    return os.path.join(schemas_path, "test_schema_samples.yaml")


@pytest.fixture
def schema_invalid_file_path(schemas_path):
    return os.path.join(schemas_path, "test_schema_invalid.yaml")


@pytest.fixture
def schema_sample_invalid_file_path(schemas_path):
    return os.path.join(schemas_path, "test_schema_sample_invalid.yaml")


@pytest.fixture
def schema_imports_file_path(schemas_path):
    return os.path.join(schemas_path, "test_schema_imports.yaml")


@pytest.fixture
def schema_rel_path_imports_file_path(schemas_path):
    return os.path.join(schemas_path, "test_schema_imports_rel_path.yaml")


@pytest.fixture
def taxprofiler_project_path(peps_path):
    return os.path.join(peps_path, "multiline_output", "config.yaml")


@pytest.fixture
def taxprofiler_project(taxprofiler_project_path):
    return Project(taxprofiler_project_path)


@pytest.fixture
def path_to_taxprofiler_csv_multiline_output(peps_path):
    return os.path.join(peps_path, "multiline_output", "multiline_output.csv")


@pytest.fixture
def path_pep_with_fasta_column(peps_path):
    return os.path.join(peps_path, "pep_with_fasta_column", "config.yaml")


@pytest.fixture
def project_pep_with_fasta_column(path_pep_with_fasta_column):
    return Project(path_pep_with_fasta_column, sample_table_index="sample")


@pytest.fixture
def output_pep_with_fasta_column(path_pep_with_fasta_column):
    with open(
        os.path.join(os.path.dirname(path_pep_with_fasta_column), "output.csv")
    ) as f:
        return f.read()


@pytest.fixture
def taxprofiler_csv_multiline_output(path_to_taxprofiler_csv_multiline_output):
    with open(path_to_taxprofiler_csv_multiline_output, "r") as file:
        data = file.read()
    return data
    # This is broken unless I add na_filter=False. But it's a bad idea anyway, since
    # we're just using this for string comparison anyway...
    # return pd.read_csv(
    #     path_to_taxprofiler_csv_multiline_output, na_filter=False
    # ).to_csv(path_or_buf=None, index=None)


@pytest.fixture
def path_pep_for_schema_with_rel_path(peps_path):
    return os.path.join(peps_path, "pep_schema_rel_path", "config.yaml")


@pytest.fixture
def path_pep_nextflow_taxprofiler(peps_path):
    return os.path.join(peps_path, "pep_nextflow_taxprofiler", "config.yaml")


@pytest.fixture
def project_pep_nextflow_taxprofiler(path_pep_nextflow_taxprofiler):
    return Project(path_pep_nextflow_taxprofiler, sample_table_index="sample")


@pytest.fixture
def output_pep_nextflow_taxprofiler(path_pep_nextflow_taxprofiler):
    with open(
        os.path.join(os.path.dirname(path_pep_nextflow_taxprofiler), "output.csv")
    ) as f:
        return f.read()


@pytest.fixture
def save_result_mock(mocker):
    return mocker.patch("peppy.eido.conversion.save_result")


@pytest.fixture
def test_file_existing_schema(schemas_path):
    return os.path.join(schemas_path, "schema_test_file_exist.yaml")


@pytest.fixture
def test_file_existing_pep(peps_path):
    return os.path.join(peps_path, "test_file_existing", "project_config.yaml")


@pytest.fixture
def test_schema_value_check(schemas_path):
    return os.path.join(schemas_path, "value_check_schema.yaml")


@pytest.fixture
def test_file_value_check(peps_path):
    return os.path.join(peps_path, "value_check_pep", "project_config.yaml")


@pytest.fixture
def test_multiple_subs(peps_path):
    return os.path.join(peps_path, "multiple_subsamples", "project_config.yaml")



from yaml import safe_load

from peppy.eido.schema import read_schema


class TestSchemaReading:
    def test_imports_file_schema(self, schema_imports_file_path):
        s = read_schema(schema_imports_file_path)
        assert isinstance(s, list)
        assert len(s) == 2

    def test_imports_dict_schema(self, schema_imports_file_path):
        with open(schema_imports_file_path, "r") as f:
            s = read_schema(safe_load(f))
        assert isinstance(s, list)
        assert len(s) == 2



import os

import pytest


@pytest.fixture
def SAMPLE_PEP():
    return os.path.join(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
        "data",
        "phcdata",
        "sample_pep",
        "subsamp_config.yaml",
    )


@pytest.fixture()
def test_raw_pep_return():
    sample_prj = {
        "config": {
            "This": "is config",
            "description": "desc",
            "name": "sample name",
        },
        "subsample_list": [],
        "sample_list": [
            {"time": "0", "file_path": "source1", "sample_name": "pig_0h"},
            {"time": "1", "file_path": "source1", "sample_name": "pig_1h"},
            {"time": "0", "file_path": "source1", "sample_name": "frog_0h"},
        ],
    }
    return sample_prj


@pytest.fixture
def requests_get_mock(mocker):
    return mocker.patch("requests.get")


@pytest.fixture
def input_return_mock(monkeypatch):
    return monkeypatch.setattr("builtins.input", lambda: None)


@pytest.fixture
def test_jwt():
    return (
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
        "eyJsb2dpbiI6InJhZmFsc3RlcGllbiIsImlkIjo0MzkyNjUyMiwib3JnYW5pemF0aW9ucyI6bnVsbH0."
        "mgBP-7x5l9cqufhzdVi0OFA78pkYDEymwPFwud02BAc"
    )



"""Helpers without an obvious logical home."""

import logging
import os
import posixpath as psp
import re
from collections import defaultdict
from collections.abc import Mapping
from typing import Any
from urllib.request import urlopen

import yaml
from ubiquerg import expandpath, is_url

from .const import CONFIG_KEY, SAMPLE_TABLE_INDEX_KEY, SUBSAMPLE_TABLE_INDEX_KEY
from .exceptions import RemoteYAMLError

_LOGGER = logging.getLogger(__name__)


def copy(obj: Any) -> Any:
    def copy(self):
        """
        Copy self to a new object.
        """
        from copy import deepcopy

        return deepcopy(self)

    obj.copy = copy
    return obj


def make_abs_via_cfg(
    maybe_relpath: str, cfg_path: str, check_exists: bool = False
) -> str:
    """
    Ensure that a possibly relative path is absolute.

    Args:
        maybe_relpath: Path that may be relative
        cfg_path: Path to configuration file
        check_exists: Whether to verify the resulting path exists

    Returns:
        Absolute path

    Raises:
        TypeError: If maybe_relpath is not a string
        OSError: If check_exists is True and path doesn't exist
    """
    if not isinstance(maybe_relpath, str):
        raise TypeError(
            "Attempting to ensure non-text value is absolute path: {} ({})".format(
                maybe_relpath, type(maybe_relpath)
            )
        )
    if os.path.isabs(maybe_relpath) or is_url(maybe_relpath):
        _LOGGER.debug("Already absolute")
        return maybe_relpath
    # Maybe we have env vars that make the path absolute?
    expanded = expandpath(maybe_relpath)
    if os.path.isabs(expanded):
        _LOGGER.debug("Expanded: {}".format(expanded))
        return expanded
    # Set path to an absolute path, relative to project config.
    if is_url(cfg_path):
        config_dirpath = psp.dirname(cfg_path)
    else:
        config_dirpath = os.path.dirname(cfg_path)
    _LOGGER.debug("config_dirpath: {}".format(config_dirpath))

    if is_url(cfg_path):
        abs_path = psp.join(config_dirpath, maybe_relpath)
    else:
        abs_path = os.path.join(config_dirpath, maybe_relpath)
    _LOGGER.debug("Expanded and/or made absolute: {}".format(abs_path))
    if check_exists and not is_url(abs_path) and not os.path.exists(abs_path):
        raise OSError(f"Path made absolute does not exist: {abs_path}")
    return abs_path


def grab_project_data(prj: Any) -> Mapping:
    """
    From the given Project, grab Sample-independent data.

    There are some aspects of a Project of which it's beneficial for a Sample
    to be aware, particularly for post-hoc analysis. Since Sample objects
    within a Project are mutually independent, though, each doesn't need to
    know about any of the others. A Project manages its Sample instances,
    so for each Sample knowledge of Project data is limited. This method
    facilitates adoption of that conceptual model.

    Args:
        prj: Project from which to grab data

    Returns:
        Sample-independent data sections from given Project

    Raises:
        KeyError: If project lacks required config section
    """
    if not prj:
        return {}

    try:
        return dict(prj[CONFIG_KEY])
    except KeyError:
        raise KeyError("Project lacks section '{}'".format(CONFIG_KEY))


def make_list(arg: list | str, obj_class: type) -> list:
    """
    Convert an object of predefined class to a list or ensure list contains correct type.

    Args:
        arg: Object or list of objects to listify
        obj_class: Class that objects should be instances of

    Returns:
        List of objects of the predefined class

    Raises:
        TypeError: If a faulty argument was provided
    """

    def _raise_faulty_arg():
        raise TypeError(
            "Provided argument has to be a List[{o}] or a {o}, got '{a}'".format(
                o=obj_class.__name__, a=arg.__class__.__name__
            )
        )

    if isinstance(arg, obj_class):
        return [arg]
    elif isinstance(arg, list):
        if not all(isinstance(i, obj_class) for i in arg):
            _raise_faulty_arg()
        else:
            return arg
    else:
        _raise_faulty_arg()


def _expandpath(path: str) -> str:
    """
    Expand a filesystem path that may or may not contain user/env vars.

    Args:
        path: Path to expand

    Returns:
        Expanded version of input path
    """
    return os.path.expandvars(os.path.expanduser(path))


def expand_paths(x: dict) -> dict:
    """
    Recursively expand paths in a dict.

    Args:
        x: dict to expand

    Returns:
        dict with expanded paths
    """
    if isinstance(x, str):
        return expandpath(x)
    elif isinstance(x, Mapping):
        return {k: expand_paths(v) for k, v in x.items()}
    return x


def load_yaml(filepath: str) -> dict:
    """
    Load a local or remote YAML file into a Python dict.

    Args:
        filepath: Path to the file to read

    Returns:
        Read data

    Raises:
        RemoteYAMLError: If the remote YAML file reading fails
    """
    if is_url(filepath):
        _LOGGER.debug(f"Got URL: {filepath}")
        try:
            response = urlopen(filepath)
        except Exception as e:
            raise RemoteYAMLError(
                f"Could not load remote file: {filepath}. "
                f"Original exception: {getattr(e, 'message', repr(e))}"
            )
        else:
            data = response.read().decode("utf-8")
            return expand_paths(yaml.safe_load(data))
    else:
        with open(os.path.abspath(filepath), "r") as f:
            data = yaml.safe_load(f)
        return expand_paths(data)


def is_cfg_or_anno(file_path: str | None, formats: dict | None = None) -> bool | None:
    """
    Determine if the input file seems to be a project config file (based on extension).

    Args:
        file_path: File path to examine
        formats: Formats dict to use. Must include 'config' and 'annotation' keys

    Returns:
        True if the file is a config, False if the file is an annotation,
        None if file_path is None

    Raises:
        ValueError: If the file seems to be neither a config nor an annotation
    """
    formats_dict = formats or {
        "config": (".yaml", ".yml"),
        "annotation": (".csv", ".tsv"),
    }
    if file_path is None:
        return None
    if file_path.lower().endswith(formats_dict["config"]):
        _LOGGER.debug(f"Creating a Project from a YAML file: {file_path}")
        return True
    elif file_path.lower().endswith(formats_dict["annotation"]):
        _LOGGER.debug(f"Creating a Project from a CSV file: {file_path}")
        return False
    raise ValueError(
        f"File path '{file_path}' does not point to an annotation or config. "
        f"Accepted extensions: {formats_dict}"
    )


def extract_custom_index_for_sample_table(pep_dictionary: dict) -> str | None:
    """
    Extracts a custom index for the sample table if it exists.

    Args:
        pep_dictionary: PEP configuration dictionary

    Returns:
        Custom index name or None if not specified
    """
    return (
        pep_dictionary[SAMPLE_TABLE_INDEX_KEY]
        if SAMPLE_TABLE_INDEX_KEY in pep_dictionary
        else None
    )


def extract_custom_index_for_subsample_table(pep_dictionary: dict) -> str | None:
    """
    Extracts a custom index for the subsample table if it exists.

    Args:
        pep_dictionary: PEP configuration dictionary

    Returns:
        Custom index name or None if not specified
    """
    return (
        pep_dictionary[SUBSAMPLE_TABLE_INDEX_KEY]
        if SUBSAMPLE_TABLE_INDEX_KEY in pep_dictionary
        else None
    )


def unpopulated_env_var(paths: set[str]) -> None:
    """
    Print warnings for unpopulated environment variables in paths.

    Given a set of paths that may contain env vars, group by env var and
    print a warning for each group with the deepest common directory and
    the paths relative to that directory.

    Args:
        paths: Set of paths that may contain environment variables
    """
    _VAR_RE = re.compile(r"^\$(\w+)/(.*)$")
    groups: dict[str, list[str]] = defaultdict(list)

    # 1) Group by env var
    for s in paths:
        m = _VAR_RE.match(s.strip())
        if not m:
            # Not in "$VAR/..." form — skip or collect under a special key if you prefer
            continue
        var, tail = m.group(1), m.group(2)
        # normalize to POSIX-ish, no leading "./"
        tail = tail.lstrip("/")
        groups[var].append(tail)

    # 2) For each var, compute deepest common directory and print
    for var, tails in groups.items():
        if not tails:
            continue

        if len(tails) == 1:
            # With a single path, use its directory as the common dir
            common_dir = psp.dirname(tails[0]) or "."
        else:
            common_dir = psp.commonpath(tails) or "."
            # Ensure it's a directory; commonpath is component-wise, so it's fine.

        warning_message = "Not all environment variables were populated in derived attribute source: $%s/{"

        in_env = []
        for t in tails:
            rel = psp.relpath(t, start=common_dir or ".")
            in_env.append(rel)

        warning_message += ", ".join(in_env)
        warning_message += "}"
        _LOGGER.warning(
            warning_message,
            var,
        )



class BasePephubclientException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class IncorrectQueryStringError(BasePephubclientException):
    def __init__(self, query_string: str | None = None):
        self.query_string = query_string
        super().__init__(
            f"PEP data with passed namespace and project ({self.query_string}) name not found."
        )


class ResponseError(BasePephubclientException):
    default_message = "The response looks incorrect and must be verified manually."

    def __init__(self, message: str | None = None):
        self.message = message
        super().__init__(self.message or self.default_message)


class PEPExistsError(BasePephubclientException):
    default_message = (
        "PEP already exists. Change location, delete previous PEP, or set force argument "
        "to overwrite previous PEP"
    )

    def __init__(self, message: str | None = None):
        self.message = message
        super().__init__(self.message or self.default_message)



import logging

from ..constants import PEPHUB_SAMPLE_URL, ResponseStatusCodes
from ..exceptions import ResponseError
from ..helpers import RequestManager

_LOGGER = logging.getLogger("pephubclient")


class PEPHubSample(RequestManager):
    """
    Class for managing samples in PEPhub.

    Provides methods for getting, creating, updating and removing samples.
    This class is not related to peppy.Sample class.
    """

    def __init__(self, jwt_data: str = None):
        """
        Initialize PEPHubSample.

        Args:
            jwt_data: jwt token for authorization
        """

        self.__jwt_data = jwt_data

    def get(
        self,
        namespace: str,
        name: str,
        tag: str,
        sample_name: str = None,
    ) -> dict:
        """
        Get sample from project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            sample_name: sample name

        Returns:
            Sample object.
        """
        url = self._build_sample_request_url(
            namespace=namespace, name=name, sample_name=sample_name
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="GET", url=url, headers=self.parse_header(self.__jwt_data)
        )
        if response.status_code == ResponseStatusCodes.OK:
            return self.decode_response(response, output_json=True)
        if response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError(
                f"Sample does not exist. Project: '{namespace}/{name}:{tag}'. Sample_name: '{sample_name}'"
            )
        elif response.status_code == ResponseStatusCodes.INTERNAL_ERROR:
            raise ResponseError("Internal server error. Unexpected return value.")
        else:
            raise ResponseError(
                f"Unexpected return value. Error: {response.status_code}"
            )

    def create(
        self,
        namespace: str,
        name: str,
        tag: str,
        sample_name: str,
        sample_dict: dict,
        overwrite: bool = False,
    ) -> None:
        """
        Create sample in project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            sample_dict: sample dict
            sample_name: sample name
            overwrite: overwrite sample if it exists
        """
        url = self._build_sample_request_url(
            namespace=namespace,
            name=name,
            sample_name=sample_name,
        )

        url = url + self.parse_query_param(
            pep_variables={"tag": tag, "overwrite": overwrite}
        )

        # add sample name to sample_dict if it is not there
        if sample_name not in sample_dict.values():
            sample_dict["sample_name"] = sample_name

        response = self.send_request(
            method="POST",
            url=url,
            headers=self.parse_header(self.__jwt_data),
            json=sample_dict,
        )
        if response.status_code == ResponseStatusCodes.ACCEPTED:
            _LOGGER.info(
                f"Sample '{sample_name}' added to project '{namespace}/{name}:{tag}' successfully."
            )
            return None
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError(f"Project '{namespace}/{name}:{tag}' does not exist.")
        elif response.status_code == ResponseStatusCodes.CONFLICT:
            raise ResponseError(
                f"Sample '{sample_name}' already exists. Set overwrite to True to overwrite sample."
            )
        else:
            raise ResponseError(
                f"Unexpected return value. Error: {response.status_code}"
            )

    def update(
        self,
        namespace: str,
        name: str,
        tag: str,
        sample_name: str,
        sample_dict: dict,
    ):
        """
        Update sample in project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            sample_name: sample name
            sample_dict: sample dict, that contain elements to update
        """

        url = self._build_sample_request_url(
            namespace=namespace, name=name, sample_name=sample_name
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="PATCH",
            url=url,
            headers=self.parse_header(self.__jwt_data),
            json=sample_dict,
        )
        if response.status_code == ResponseStatusCodes.ACCEPTED:
            _LOGGER.info(
                f"Sample '{sample_name}' updated in project '{namespace}/{name}:{tag}' successfully."
            )
            return None
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError(
                f"Sample '{sample_name}' or project {namespace}/{name}:{tag} does not exist. Error: {response.status_code}"
            )
        else:
            raise ResponseError(
                f"Unexpected return value. Error: {response.status_code}"
            )

    def remove(self, namespace: str, name: str, tag: str, sample_name: str):
        """
        Remove sample from project in PEPhub.

        Args:
            namespace: namespace of project
            name: name of project
            tag: tag of project
            sample_name: sample name
        """
        url = self._build_sample_request_url(
            namespace=namespace, name=name, sample_name=sample_name
        )

        url = url + self.parse_query_param(pep_variables={"tag": tag})

        response = self.send_request(
            method="DELETE",
            url=url,
            headers=self.parse_header(self.__jwt_data),
        )
        if response.status_code == ResponseStatusCodes.ACCEPTED:
            _LOGGER.info(
                f"Sample '{sample_name}' removed from project '{namespace}/{name}:{tag}' successfully."
            )
            return None
        elif response.status_code == ResponseStatusCodes.NOT_EXIST:
            raise ResponseError(
                f"Sample '{sample_name}' or project {namespace}/{name}:{tag} does not exist. Error: {response.status_code}"
            )
        else:
            raise ResponseError(
                f"Unexpected return value. Error: {response.status_code}"
            )

    @staticmethod
    def _build_sample_request_url(namespace: str, name: str, sample_name: str) -> str:
        """
        Build url for sample request.

        Args:
            namespace: namespace where project will be uploaded

        Returns:
            url string.
        """
        return PEPHUB_SAMPLE_URL.format(
            namespace=namespace, project=name, sample_name=sample_name
        )
